/*-
 * SPDX-License-Identifier: LGPL-2.1-only
 * Copyright 2022,2023 UPLEX Nils Goroll Systemoptimierung. All rights reserved.
 * Author: Nils Goroll <nils.goroll@uplex.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301 USA Also add information on how to contact you by
 * electronic and paper mail.
 */

#include "config.h"

#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>

#include "cache/cache_varnishd.h"

#include "debug.h"
#include "compiler.h"
#include "buddy.h"
#include "fellow_pri.h"
#include "fellow_io.h"
#include "fellow_io_backend.h"
#include "fellow_log_storage.h"
#include "fellow_log.h"
#include "fellow_hash.h"
#include "fellow_cache.h"
#include "fellow_cache_storage.h"
#include "fellow_tune.h"

#ifdef TEST_DRIVER
#define FCERR_INJECT
#endif
#include "fellow_inject.h"

static char wrongbuf[1024];

#define FC_WRONG(...) do {			\
		bprintf(wrongbuf, __VA_ARGS__);	\
		WRONG(wrongbuf);		\
	} while (0)

struct fellow_cache_obj;
// look up objects by fdb
VRBT_HEAD(fellow_cache_fdb_head, fellow_cache_obj);

struct fellow_cache_seg;
VTAILQ_HEAD(fellow_cache_lru_head, fellow_cache_seg);

/*
 * lifetime of a fellow_obj in cache
 *
 * presence in fdb tree is reliable
 * priv PTR is hint
 *
 * 1 create fellow_cache_obj and ref in fdb tree
 * 2 fetch obj
 *
 *
 * remove from cache:
 * - nuke  : varnish-cache forgets about the object
 * - evict : object still exists, but removing from memory
 *
 * eviction:
 * - remove stobj priv ptr
 * - remove from fdb tree
 * - remove from lru
 * when refcnt == 0:
 * - free
 *
 * refcnt'ing obj (fco) and body (fcs)
 *
 * IF body is needed:
 * - for fco INCORE, grab ref on all body fcs
 *   for fco FETCHING -> INCORE this happens with the transition
 *
 * - fco ref must be held if any body fcs ref is held
 */

enum fcos {
	FCOS_INIT = 1,
	FCOS_USABLE,
	FCOS_BUSY,
	FCOS_WRITING,
	FCOS_DISK,
	FCOS_MEM,
	FCOS_READING,
	FCOS_CHECK,
	FCOS_REDUNDANT,
	FCOS_INCORE,
	FCOS_READFAIL,
	FCOS_EVICT,
	FCOS_MAX
} __attribute__ ((__packed__));

#define FCOS_SHIFT 4
#define FCOS_MASK ((1 << FCOS_SHIFT) - 1)
#define FCO_HIGH (2 << FCOS_SHIFT)
#define FCS_HIGH (3 << FCOS_SHIFT)
#define FCOS_HIGH(x) (x & FCS_HIGH)

#define FCO(fcos) (FCO_HIGH | FCOS_ ## fcos)
#define FCS(fcos) (FCS_HIGH | FCOS_ ## fcos)
#define FCOS(x) (x & FCOS_MASK)

#define FCOSD(x, y, c, lru, fdb, mem, dsk) x ## _ ## y = x(y),

//lint -e{123} macro FCO
enum fcos_state {
	FCOS_INVAL = 0,
#include "tbl/fco_states.h"
	FCO_MAX,
#include "tbl/fcs_states.h"
	FCS_MAX,
	FCOS_STATE_MAX = FCS_MAX
} __attribute__ ((__packed__));
#undef FCOSD
//lint -e{786} string concat
//lint -e{785} too few
static const char * const fcos_state_s[FCOS_STATE_MAX] = {
#define FCOSD(x, y, c, lru, fdb, mem, dsk)  [x ## _ ## y] = #x "_" #y,
#include "tbl/fco_states.h"
#include "tbl/fcs_states.h"
};
#undef FCOSD

#define FCOS_IS(state, fcos) (FCOS(state) == FCOS_ ## fcos)

static void __attribute__((constructor))
assert_fcos(void)
{
	//lint --e{506} constant boolean
	assert(FCOS_MAX <= FCOS_MASK);
	assert(FCO_READFAIL == FCO(READFAIL));
	assert(FCS_READFAIL == FCS(READFAIL));
	assert(FCOS_IS(FCS_INCORE, INCORE));
	assert(FCOS_IS(FCO_INCORE, INCORE));
	assert(! FCOS_IS(FCO_INCORE, DISK));
}

#define FCOS_BIT(fcos) ((uint16_t)1 << (fcos - 1))

// this is enum fcos_state -> FCOS_BIT(enum fcos)
static uint16_t fcos_transition[FCOS_STATE_MAX];

static void __attribute__((constructor))
init_fcos_transitions(void)
{

	memset(fcos_transition, 0, sizeof fcos_transition);
#define FCOT(f, t, c)						\
	fcos_transition[FCO_ ## f] |= FCOS_BIT(FCOS_ ## t);
#include "tbl/fco_transition.h"
#define FCST(f, t, c)						\
	fcos_transition[FCS_ ## f] |= FCOS_BIT(FCOS_ ## t);
#include "tbl/fcs_transition.h"
}

static inline void
assert_fcos_transition(enum fcos_state f, enum fcos_state t)
{

	assert(FCOS_HIGH(f) == FCOS_HIGH(t));
	if (! (fcos_transition[f] & FCOS_BIT(FCOS(t))))
		FC_WRONG("transition %s -> %s",
		    fcos_state_s[f], fcos_state_s[t]);
}

static inline void
fellow_cache_seg_transition_locked_notincore(struct fellow_cache_seg *fcs,
    enum fcos_state to);

static inline void
fellow_cache_seg_transition_locked(
    struct fellow_cache *fc, struct fellow_cache_seg *fcs,
    enum fcos_state from, enum fcos_state to);

struct fellow_disk_seg {
	uint16_t			magic;
#define FELLOW_DISK_SEG_MAGIC		0xf93d
	uint8_t				_reserved;
	uint8_t				fht;
	uint32_t			segnum;
	struct buddy_off_extent	seg;

	// checksum of data referred to by seg
	// EXCEPT for fellow_disk_obj where it is only
	// the bottom of fellow_disk_obj + va_data
	union fh			fh[1];
};

struct fellow_disk_seglist {
	uint32_t			magic;
#define FELLOW_DISK_SEGLIST_MAGIC	0x06bbf521
	uint8_t				version;
	uint8_t				_reserved[2];
	uint8_t				fht;

	// off = 8 - checksum of data below
	union fh			fh[1];

	// vvv CHECKSUMMED

#define FELLOW_DISK_SEGLIST_CHK_START			\
	offsetof(struct fellow_disk_seglist, next)
#define FELLOW_DISK_SEGLIST_LEN(fdsl)				\
	(sizeof *fdsl + (fdsl)->nsegs * sizeof *(fdsl)->segs)
#define FELLOW_DISK_SEGLIST_CHK_LEN(fdsl)			\
	(FELLOW_DISK_SEGLIST_LEN(fdsl) - FELLOW_DISK_SEGLIST_CHK_START)

	struct buddy_off_extent	next;
	uint16_t			nsegs;
	uint16_t			lsegs;
	uint8_t				_reserved2[4];
	struct fellow_disk_seg		segs[];
};

#define fellow_disk_seglist_size(fdsl, lsegs)		\
	(sizeof *fdsl + lsegs * sizeof *fdsl->segs)

struct fellow_cache_obj;
struct fellow_cache_seg {
	uint16_t			magic;
#define FELLOW_CACHE_SEG_MAGIC		0x6278
	enum fcos_state			state;
	unsigned			fcs_onlru:1;
	unsigned			fco_infdb:1;

	/*
	 * for FCO, protected by fdb_mtx
	 * for FCS, protected by fco mtx
	 */
	unsigned			refcnt;

	VTAILQ_ENTRY(fellow_cache_seg)	lru_list;
	struct fellow_cache_obj	*fco;
	struct fellow_disk_seg		*disk_seg;
	struct buddy_ptr_extent	alloc;
	size_t				len;	// used size
};

struct fellow_cache_seglist {
	uint32_t			magic;
#define FELLOW_CACHE_SEGLIST_MAGIC	0xcad6e9db
	uint16_t			lsegs;		// how many av.

	struct fellow_disk_seglist	*fdsl;		// to free
	size_t				fdsl_sz;	// 0 for embedded
	size_t				fcsl_sz;	// 0 for embedded

	struct fellow_cache_seglist	*next;
	struct fellow_cache_seg	segs[];
};

#define fellow_cache_seglist_size(fcsl, lsegs)		\
	(sizeof *fcsl + lsegs * sizeof *fcsl->segs)

struct fellow_disk_obj_attr {
	uint32_t	aoff;	// from struct fellow_disk_obj
	uint32_t	alen;
};

struct fcscursor {
	struct fellow_cache_seglist *fcsl;
	struct fellow_disk_seglist *fdsl;
	unsigned u;
};

static inline void
fcsc_init(struct fcscursor *c, struct fellow_cache_seglist *fcsl)
{
	CHECK_OBJ_NOTNULL(fcsl, FELLOW_CACHE_SEGLIST_MAGIC);
	c->fcsl = fcsl;
	c->fdsl = fcsl->fdsl;
	CHECK_OBJ_NOTNULL(c->fdsl, FELLOW_DISK_SEGLIST_MAGIC);
	c->u = 0;
}

static inline struct fellow_cache_seg *
fcsc_next(struct fcscursor *c)
{
	if (c->u == c->fdsl->nsegs) {
		if (c->fcsl->next == NULL)
			return (NULL);
		fcsc_init(c, c->fcsl->next);
	}
	assert(c->u < c->fdsl->nsegs);
	return (&c->fcsl->segs[c->u++]);
}

// fdo_flags
#define FDO_F_INLOG	1

/*
 * disk layout
 *
 * struct fellow_disk_obj	disk_obj
 * uint8_t			varattr_data[variable]
 * fellow_disk_seglist		seglist
 */

struct fellow_disk_obj {
	uint32_t			magic;
#define FELLOW_DISK_OBJ_MAGIC		0x50728fbd
	uint8_t			version;
	uint8_t			_reserved[3];

	// segment off/sz is for fellow_disk_obj + seglist
	// checksum is only for data below
	struct fellow_disk_seg		fdo_fds;
#define FELLOW_DISK_OBJ_CHK_START		\
	offsetof(struct fellow_disk_obj, va_data_len)
#define FELLOW_DISK_OBJ_CHK_LEN(fdo)					\
	((sizeof(struct fellow_disk_obj) - FELLOW_DISK_OBJ_CHK_START) +	\
	(fdo)->va_data_len)

	/* FDO_FIXATTR is 2, 8, 8, 8, 32
	 * so we use 4, 1, 1 to pad to 2x 32 */
	uint32_t			va_data_len;
	uint8_t			fdo_flags;
	uint8_t			fdo_reserve;

	/* attributes */
#define FDO_FIXATTR(U, l, s, vs)			\
	uint8_t			fa_##l[s];
#include "tbl/fellow_obj_attr.h"
	uint8_t			fa_reserve[64];

#define FDO_VARATTR(U, l)			\
	struct fellow_disk_obj_attr	va_##l;
#include "tbl/fellow_obj_attr.h"
	struct fellow_disk_obj_attr	va_reserve[4];

#define FDO_AUXATTR(U, l)			\
	struct fellow_disk_seg		aa_##l##_seg;
#include "tbl/fellow_obj_attr.h"
	// no need for aa_reserve, va offset is variable

	uint8_t			va_data[];
};

// shaved off one pointer...
#define FCO_FCS(fco)	(&(fco)->fdo_fcs)
#define FCO_FDO(fco)	(void*)(fco)->fdo_fcs.alloc.ptr
#define FCO_REFCNT(fco) (fco)->fdo_fcs.refcnt
#define FCO_STATE(fco)	(fco)->fdo_fcs.state

/*
 * dunno -> wantlog
 * dunno -> nolog
 * dunno -> toolate
 * dunno -> inlog	// read from disk
 * wantlog -> toolate
 * wantlog -> inlog
 * inlog -> deleted
 */
enum fcol_state {
	FCOL_DUNNO = 0,	// not yet decided
	FCOL_WANTLOG,
	FCOL_NOLOG,	// private object
	FCOL_TOOLATE,	// already dead
	FCOL_INLOG,
	FCOL_DELETED,
	FCOL__MAX
};

struct fellow_cache_obj {
	unsigned			magic;
#define FELLOW_CACHE_OBJ_MAGIC		0x837d555f
	// protected by fco mtx (in fcs)
	enum fcol_state		logstate;

	struct buddy_ptr_page		fco_mem;	// nil == embedded

	pthread_mutex_t		mtx;
	pthread_cond_t			cond;	// signal ready

	struct objcore			*oc;

	fellow_disk_block		fdb;
	VRBT_ENTRY(fellow_cache_obj)	fdb_entry;

	struct fellow_cache_seg		fdo_fcs;
#define FDO_AUXATTR(U, l)				\
	struct fellow_cache_seg		aa_##l##_seg;
#include "tbl/fellow_obj_attr.h"

	struct fellow_cache_seglist	seglist;
};

#define FCR_OK(p)\
	(struct fellow_cache_res){.r.ptr = (p), .status = fcr_ok}
#define FCR_ALLOCERR(str)\
	(struct fellow_cache_res){.r.err = (str), .status = fcr_allocerr}
#define FCR_IOERR(str)\
	(struct fellow_cache_res){.r.err = (str), .status = fcr_ioerr}
#define FCR_ALLOCFAIL(str) FCR_ALLOCERR(FC_ERRSTR(str))
#define FCR_IOFAIL(str) FCR_IOERR(FC_ERRSTR(str))

const char * const fellow_cache_res_s[FCR_LIM] = {
	[fcr_ok] = "ok",
	[fcr_allocerr] = "allocation",
	[fcr_ioerr] = "io"
};

struct fellow_body_region {
	struct buddy_off_extent	*reg;
	// used space
	size_t				len;
	// counting allocations from the same region
	unsigned			segnum;
};

/* info struct for busy writes */
enum fellow_busy_io_e {
	FBIO_INVAL = 0,
	FBIO_SEG,
	FBIO_SEGLIST
} __attribute__ ((__packed__));

enum fellow_busy_iosync_e {
	FBIOS_INVAL = 0,
	FBIOS_SYNC,
	FBIOS_ASYNC
} __attribute__ ((__packed__));

// XXX tunable?
#define FBIO_MAX_RETRIES 3

struct fellow_busy_io_seglist {
	struct fellow_disk_seglist *fdsl;
	struct buddy_off_extent reg;
};

struct fellow_busy_io {
	uint16_t				magic;
#define FELLOW_BUSY_IO_MAGIC			0x0bcb
	uint16_t				retries;
	enum fellow_busy_io_e			type;
	enum fellow_busy_iosync_e		sync;
	struct fellow_busy			*fbo;
	union {
		// for FBIO_SEG
		struct fellow_cache_seg	*fcs;
		// for FBIO_SEGLIST
		struct fellow_busy_io_seglist	seglist;
	} u;
};

// stored in stevedore priv
struct fellow_busy {
	unsigned			magic;
#define FELLOW_BUSY_MAGIC		0x8504a132

	struct buddy_ptr_extent	fbo_mem;
	size_t				sz_estimate;

	// sum of getspace returned
	size_t				sz_returned;
	// we add more whenever we exceed estimate
	size_t				sz_increment;

	// how much we already alloced on dsk
	size_t				sz_dskalloc;

	struct fellow_cache		*fc;
	struct fellow_cache_obj	*fco;
	struct fellow_cache_seglist	*body_seglist;
	struct fellow_cache_seg	*body_seg;
	struct fellow_body_region	body_region;

#define FCO_MAX_REGIONS ((FELLOW_DISK_LOG_BLOCK_ENTRIES - 1) * DLE_REG_NREGION)
	struct buddy_off_extent	region[FCO_MAX_REGIONS];
	unsigned			nregion;

	/* protected by fco mtx */
	uint8_t				io_idx;
	uint8_t				io_outstanding;
	struct fellow_busy_io		io[FCO_MAX_REGIONS];

	// for fellow_busy_setattr()
	uint32_t			va_data_len;
	uint8_t			*va_data;
};

struct fellow_cache {
	unsigned			magic;
#define FELLOW_CACHE_MAGIC		0xe2f2243e
	unsigned			running;

	buddy_t				*membuddy;
	struct fellow_fd		*ffd;
	struct stvfe_tune		*tune;

	pthread_mutex_t		fdb_mtx;
	struct fellow_cache_fdb_head	fdb_head;

	pthread_mutex_t		lru_mtx;
	struct fellow_cache_lru_head	lru_head;

	pthread_mutex_t		async_mtx;
	pthread_cond_t			async_cond;
	void				*async_ioctx;
	pthread_t			async_thread;
	unsigned			async_idle;
};

/* ============================================================
 * util
 */

static void
fellow_cache_res_check(const struct fellow_cache *fc,
    const struct fellow_cache_res fcr)
{
	struct stvfe_tune *tune;

	CHECK_OBJ_NOTNULL(fc, FELLOW_CACHE_MAGIC);
	tune = fc->tune;
	CHECK_OBJ_NOTNULL(tune, STVFE_TUNE_MAGIC);

	switch (fcr.status) {
	case fcr_allocerr:
		if (tune->allocerr_obj)
			return;
		break;
	case fcr_ioerr:
		if (tune->ioerr_obj)
			return;
		break;
	case fcr_ok:
	default:
		FC_WRONG("fcr.status %d", fcr.status);
	}

	FC_WRONG("fcr.status = %d, fcr.r.err = %s",
	    fcr.status, fcr.r.err);
}


/* ============================================================
 * fwd decl
 */
static void
fellow_cache_obj_fini(const struct fellow_cache_obj *fco);
static void
fellow_cache_obj_redundant(struct fellow_cache *fc,
    struct fellow_cache_obj **fcop);

static inline unsigned
fellow_cache_seg_ref_locked(struct fellow_cache *fc,
    struct fellow_cache_seg *fcs);
static inline unsigned
fellow_cache_seg_deref_locked(struct fellow_cache *fc,
    struct fellow_cache_seg *fcs);
static unsigned
fellow_cache_obj_deref_locked(struct fellow_cache *fc,
    struct fellow_cache_obj *fco);
static const char *
fellow_cache_seg_check(struct fellow_cache *fc, struct fellow_cache_seg *fcs);
static void
fellow_cache_async_write_complete(struct fellow_cache *fc,
    void *fbio, int32_t result);

static void
fellow_busy_log_submit(const struct fellow_busy *);
static void
fellow_busy_free(struct fellow_busy **fbop);

/* ============================================================
 * busy io
 */

/* under fco lock */
static struct fellow_busy_io *
fellow_busy_io_get(struct fellow_busy *fbo, struct fellow_busy_io *stk)
{
	struct fellow_busy_io *fbio;
	uint8_t u;

	AN(stk);

	for (u = 0; u < FCO_MAX_REGIONS; u++) {
		fbio = &fbo->io[fbo->io_idx];
		fbo->io_idx++;
		fbo->io_idx %= FCO_MAX_REGIONS;
		if (fbio->magic == 0 && fbio->type == FBIO_INVAL) {
			INIT_OBJ(fbio, FELLOW_BUSY_IO_MAGIC);
			fbio->fbo = fbo;
			fbio->sync = FBIOS_ASYNC;
			return (fbio);
		}
	}

	fbio = stk;
	INIT_OBJ(fbio, FELLOW_BUSY_IO_MAGIC);
	fbio->fbo = fbo;
	fbio->sync = FBIOS_SYNC;
	return (fbio);
}

static void
fellow_busy_io_submit(struct fellow_busy_io *fbio)
{
	struct fellow_cache_seg *fcs;
	struct fellow_disk_seg *fds;
	struct fellow_busy *fbo;
	struct fellow_cache *fc;
	uint64_t info;
	void *ptr;
	size_t size;
	off_t off;
	int r;

	CHECK_OBJ_NOTNULL(fbio, FELLOW_BUSY_IO_MAGIC);
	fbo = fbio->fbo;
	CHECK_OBJ_NOTNULL(fbo, FELLOW_BUSY_MAGIC);
	fc = fbo->fc;
	CHECK_OBJ_NOTNULL(fc, FELLOW_CACHE_MAGIC);

	switch (fbio->type) {
	case FBIO_SEG:
		fcs = fbio->u.fcs;
		CHECK_OBJ_NOTNULL(fcs, FELLOW_CACHE_SEG_MAGIC);
		fds = fcs->disk_seg;
		CHECK_OBJ_NOTNULL(fds, FELLOW_DISK_SEG_MAGIC);

		ptr  = fcs->alloc.ptr;
		size = fcs->alloc.size;
		off  = fds->seg.off;
		break;
	case FBIO_SEGLIST:
		CHECK_OBJ_NOTNULL(fbio->u.seglist.fdsl,
		    FELLOW_DISK_SEGLIST_MAGIC);
		ptr  = fbio->u.seglist.fdsl;
		size = fbio->u.seglist.reg.size;
		off  = fbio->u.seglist.reg.off;
		break;
	default:
		WRONG("fbio->type");
	}

	AN(ptr);
	AN(size);
	AN(off);	// first block is header

	switch (fbio->sync) {
	case FBIOS_ASYNC:
		info = faio_ptr_info(FAIOT_CACHE_WRITE, fbio);

		AZ(pthread_mutex_lock(&fc->async_mtx));
		r = fellow_io_write_async_enq(fc->async_ioctx,
		    info, ptr, size, off);
		if (fc->async_idle)
			AZ(pthread_cond_signal(&fc->async_cond));
		AZ(pthread_mutex_unlock(&fc->async_mtx));
		if (r)
			return;
		/* FALLTHROUGH */
	case FBIOS_SYNC:
		fellow_cache_async_write_complete(fc, fbio,
		    fellow_io_pwrite_sync(fc->ffd, ptr, size, off));
		return;
	default:
		WRONG("fbio->sync");
	}
}


/* ============================================================
 * disk obj
 */

#define fdowrong(x) do {					\
		return (FC_ERRSTR("disk object wrong " x));	\
	} while(0)

static struct fellow_disk_obj *
fellow_disk_obj(const struct fellow_cache_seg *fcs)
{
	AN(fcs);
	AN(fcs->alloc.ptr);
	return (fcs->alloc.ptr);
}

static const char *
fellow_disk_obj_check(const struct fellow_disk_obj *fdo,
    fellow_disk_block fdba)
{
	const struct fellow_disk_seg *fds;
	size_t off, len;

	AN(fdo);
	/*
	 * FC_INJ: Because the error handling path in the caller is identical we
	 * do not inject at all possible points
	 */
	if (FC_INJ || fdo->magic != FELLOW_DISK_OBJ_MAGIC)
		fdowrong("magic");
	if (fdo->version != 1)
		fdowrong("version");
	if (fdo->va_data_len + sizeof *fdo > fdb_size(fdba))
		fdowrong("va_data_len out of bounds");
	fds = &fdo->fdo_fds;
	if (FC_INJ || fds->magic != FELLOW_DISK_SEG_MAGIC)
		fdowrong("fds magic");
	if (fds->fht >= FH_LIM)
		fdowrong("hash type");
	if (fds->segnum != 0)
		fdowrong("fds segnum");
	if (fds->seg.off != fdb_off(fdba))
		fdowrong("fds off mismatch");
	if (fds->seg.size != fdb_size(fdba))
		fdowrong("fds size mismatch");
	if (FC_INJ || fhcmp(fds->fht, fds->fh,
		(uint8_t *)fdo + FELLOW_DISK_OBJ_CHK_START,
		FELLOW_DISK_OBJ_CHK_LEN(fdo)))
		fdowrong("chksum");
#define FDO_VARATTR(U, l)						\
	len = fdo->va_##l.alen;						\
	if (len > 0) {							\
		off = fdo->va_##l.aoff;					\
		if (off < offsetof(struct fellow_disk_obj, va_data))	\
			fdowrong(#l " offset");				\
		if (off + len > sizeof(*fdo) + fdo->va_data_len)	\
			fdowrong(#l " length");				\
	}
#include "tbl/fellow_obj_attr.h"

	return (NULL);
}

/* ============================================================
 * cache obj
 */

static inline int
fco_cmp(const struct fellow_cache_obj *o1, const struct fellow_cache_obj *o2)
{
	uint64_t b1, b2;

	AN(o1);
	AN(o2);
	b1 = o1->fdb.fdb;
	b2 = o2->fdb.fdb;
	// busy objects may not be added to the fdb tree
	AN(b1);
	AN(b2);

	if (b1 < b2)
		return (-1);
	if (b1 > b2)
		return (1);
	return (0);
}

VRBT_GENERATE_INSERT_COLOR(fellow_cache_fdb_head, fellow_cache_obj, fdb_entry, static)
//VRBT_GENERATE_FIND(fellow_cache_fdb_head, fellow_cache_obj, fdb_entry, fco_cmp, static)
#ifdef VRBT_GENERATE_INSERT_FINISH
VRBT_GENERATE_INSERT_FINISH(fellow_cache_fdb_head, fellow_cache_obj, fdb_entry, static)
#endif
VRBT_GENERATE_INSERT(fellow_cache_fdb_head, fellow_cache_obj, fdb_entry, fco_cmp, static)
VRBT_GENERATE_REMOVE_COLOR(fellow_cache_fdb_head, fellow_cache_obj, fdb_entry, static)
VRBT_GENERATE_REMOVE(fellow_cache_fdb_head, fellow_cache_obj, fdb_entry, static)
//VRBT_GENERATE_NEXT(fellow_cache_fdb_head, fellow_cache_obj, fdb_entry, static)
//VRBT_GENERATE_MINMAX(fellow_cache_fdb_head, fellow_cache_obj, fdb_entry, static)

static inline int
fellow_cache_shouldlru(enum fcos_state state, const struct objcore *oc,
    unsigned refcnt)
{
	if (FCOS(state) != FCOS_INCORE)
		return (0);
	if (FCOS_HIGH(state) == FCO_HIGH)
		return (oc != NULL && refcnt > 0);
	return (refcnt == 0);
}

/* chg is fellow_cache_shouldlru(new) - fellow_cache_shouldlru(old)
 *
 * iow: 0 -> noop, 1 -> add, -1 remove
 *
 * to be called after the change
 */
static inline void
fellow_cache_lru_chg(struct fellow_cache *fc, struct fellow_cache_seg *fcs,
    int chg)
{
	if (chg == 0)
		return;

	int add = chg > 0;

	assert(add ==
	    fellow_cache_shouldlru(fcs->state, fcs->fco->oc, fcs->refcnt));
	assert(fcs->fcs_onlru == (unsigned)!add);
	fcs->fcs_onlru = (unsigned)add;

	if (add) {
		AZ(pthread_mutex_lock(&fc->lru_mtx));
		VTAILQ_INSERT_TAIL(&fc->lru_head, fcs, lru_list);
		AZ(pthread_mutex_unlock(&fc->lru_mtx));
	} else {
		AZ(pthread_mutex_lock(&fc->lru_mtx));
		VTAILQ_REMOVE(&fc->lru_head, fcs, lru_list);
		AZ(pthread_mutex_unlock(&fc->lru_mtx));
	}
}

static inline void
assert_cache_seg_consistency(const struct fellow_cache_seg *fcs)
{
#define FCOSA_NOLRU AZ(fcs->fcs_onlru)
#define FCOSA_MAYLRU assert(fcs->fcs_onlru == \
	fellow_cache_shouldlru(fcs->state, fcs->fco->oc, fcs->refcnt))
#define FCOSA_NOFDB AZ(fcs->fco_infdb)
#define FCOSA_MAYFDB assert(fcs->fco_infdb == (unsigned)(fcs->refcnt > 0))
#define FCOSA_NOMEM do {				\
		AZ(fcs->alloc.ptr);			\
		AZ(fcs->alloc.size);			\
		AZ(fcs->len);				\
	} while(0)
#define FCOSA_MEM do {					\
		AN(fcs->alloc.ptr);			\
		AN(fcs->alloc.size);			\
	} while(0)
#define FDSA_NULL AZ(fcs->disk_seg)
#define FDSA_MAY CHECK_OBJ_ORNULL(fcs->disk_seg, FELLOW_DISK_SEG_MAGIC)
#define FDSA_UN do {							\
	    CHECK_OBJ_NOTNULL(fcs->disk_seg, FELLOW_DISK_SEG_MAGIC);	\
	    AZ(fcs->disk_seg->seg.off);					\
	    AZ(fcs->disk_seg->seg.size);				\
	} while(0)
#define FDSA_ALLOC do {							\
	    CHECK_OBJ_NOTNULL(fcs->disk_seg, FELLOW_DISK_SEG_MAGIC);	\
	    AN(fcs->disk_seg->seg.off);					\
	    AN(fcs->disk_seg->seg.size);				\
	} while(0)

	CHECK_OBJ_NOTNULL(fcs, FELLOW_CACHE_SEG_MAGIC);
	switch (fcs->state) {
//lint --e{525} Negative indentation
//lint --e{665} Unparan ... as expression
#define FCOSD(what, state, comment, lru, fdb, mem, fds)	\
		case what ## _ ## state:			\
			lru;					\
			fdb;					\
			mem;					\
			fds;					\
			break;
#include "tbl/fco_states.h"
#include "tbl/fcs_states.h"
	default:
		WRONG("fcs->state");
	}
#undef FCOSA_NOLRU
#undef FCOSA_MAYLRU
#undef FCOSA_NOFDB
#undef FCOSA_FDB
#undef FCOSA_NOMEM
#undef FCOSA_MEM
#undef FDSA_NULL
#undef FDSA_UN
#undef FDSA_ALLOC
}

static inline void
fellow_cache_seg_wait_locked(const struct fellow_cache_seg *fcs)
{
	AN(fcs->refcnt);
//	DBG("%p", fcs);
	assert_cache_seg_consistency(fcs);
	AZ(pthread_cond_wait(&fcs->fco->cond, &fcs->fco->mtx));
}


/*
 * free the cached segment, not the fellow_(cache|disk)_seg
 *
 * called holding the fco lock
 */
static void
fellow_cache_seg_free(struct fellow_cache *fc,
    struct fellow_cache_seg *fcs, unsigned deref)
{

	CHECK_OBJ_NOTNULL(fcs, FELLOW_CACHE_SEG_MAGIC);

	DBG("%p deref %u", fcs, deref);

	while (fcs->state == FCS_READING || fcs->state == FCS_WRITING) {
		(void) fellow_cache_seg_ref_locked(fc, fcs);
		fellow_cache_seg_wait_locked(fcs);
		deref++;
	}

	assert(
	    fcs->state == FCO_EVICT ||
	    fcs->state == FCO_REDUNDANT ||

	    fcs->state == FCS_INIT ||
	    fcs->state == FCS_USABLE ||
	    fcs->state == FCS_INCORE ||
	    fcs->state == FCS_DISK ||
	    fcs->state == FCS_CHECK ||
	    fcs->state == FCS_READFAIL);
	assert_cache_seg_consistency(fcs);
	assert(fcs->refcnt == deref);
	if (fcs->refcnt) {
		fcs->refcnt = 1;
		AZ(fellow_cache_seg_deref_locked(fc, fcs));
	}
	if (fcs->alloc.ptr) {
		buddy_return1_ptr_extent(fc->membuddy, &fcs->alloc);
		fcs->len = 0;
	}
	if (fcs->state == FCS_INCORE) {
		fellow_cache_seg_transition_locked(fc, fcs,
		    FCS_INCORE, FCS_DISK);
		assert_cache_seg_consistency(fcs);
	}
	/*
	 * at this point, the fcs is not consistent in all cases, e.g. FCS_EVICT
	 * has no memory - but this is the point where it does no longer exist
	 */
}

/*
 * references on AUXATTR are never returned after access
 * through fellow_cache_obj_getattr()
 *
 * fco->mtx held
 */
static void
fellow_cache_seg_auxattr_free(struct fellow_cache *fc,
    struct fellow_cache_seg *fcs)
{

	CHECK_OBJ_NOTNULL(fcs, FELLOW_CACHE_SEG_MAGIC);
	assert_cache_seg_consistency(fcs);
	fellow_cache_seg_free(fc, fcs, fcs->refcnt);
}

/* fco->mtx held unless surplus seglist */
static void
fellow_cache_seglist_free(struct fellow_cache *fc,
    struct fellow_cache_seglist *fcsl)
{
	struct fellow_cache_seglist *next;
	struct buddy_returns *memret;
	struct buddy_ptr_extent e;
	unsigned u, segs;

	AN(fcsl);

	memret = BUDDY_RETURNS_STK(fc->membuddy, BUDDY_RETURNS_MAX);

	while (fcsl != NULL) {
		CHECK_OBJ(fcsl, FELLOW_CACHE_SEGLIST_MAGIC);

		segs = 0;
		if (fcsl->fdsl)
			segs = fcsl->fdsl->nsegs;

		for (u = 0; u < segs; u++)
			fellow_cache_seg_free(fc, &fcsl->segs[u], 0);
		for (; u < fcsl->lsegs; u++) {
			assert(FCOS_IS(fcsl->segs[u].state, INIT) ||
			    FCOS_IS(fcsl->segs[u].state, USABLE));
			AZ(fcsl->segs[u].alloc.ptr);
		}

		if (fcsl->fdsl == NULL)
			AZ(fcsl->fdsl_sz);
		else if (fcsl->fdsl_sz > 0) {
			e = BUDDY_PTR_EXTENT(fcsl->fdsl, fcsl->fdsl_sz);
			fcsl->fdsl = NULL;
			fcsl->fdsl_sz = 0;
			AN(buddy_return_ptr_extent(memret, &e));
		}

		next = fcsl->next;
		if (fcsl->fcsl_sz > 0) {
			e = BUDDY_PTR_EXTENT(fcsl, fcsl->fcsl_sz);
			fcsl->fcsl_sz = 0;
			AN(buddy_return_ptr_extent(memret, &e));
		}

		fcsl = next;
	}
	buddy_return(memret);
}

/*
 * get back region list from segments
 */

static inline void
fellow_seg_regions(const struct fellow_fd *ffd,
    const struct fellow_cache_seg *fcs,
    struct buddy_off_extent region[FCO_MAX_REGIONS], unsigned *na)
{
	const struct fellow_disk_seg *fds;
	unsigned u, n = *na;
	buddyoff_t off;
	size_t sz;

	fds = fcs->disk_seg;
	CHECK_OBJ_NOTNULL(fds, FELLOW_DISK_SEG_MAGIC);
	sz = fellow_rndup(ffd, fds->seg.size);
	if (fds->segnum > 0) {
		for (u = 0; u < n; u++) {
			off = region[u].off + (buddyoff_t)region[u].size;
			if (off == fds->seg.off) {
				region[u].size += sz;
				assert (*na == n);
				return;
			}
		}
		WRONG("segnum > 0 needs to follow previous region");
	}
	region[n].off = fds->seg.off;
	region[n].size = sz;
	AN(fds->seg.off);
	AN(sz);
	n++;
	assert(n <= FCO_MAX_REGIONS);
	*na = n;
}

static unsigned
fellow_seglist_regions(const struct fellow_fd *ffd,
    struct fellow_cache_seglist *fcsl,
    struct buddy_off_extent region[FCO_MAX_REGIONS], unsigned n)
{
	struct fellow_cache_seg *fcs;
	struct fcscursor c;

	fcsc_init(&c, fcsl);

	while ((fcs = fcsc_next(&c)) != NULL)
		fellow_seg_regions(ffd, fcs, region, &n);
	return (n);
}

static unsigned
fellow_obj_regions(const struct fellow_fd *ffd,
    struct fellow_cache_obj *fco,
    struct buddy_off_extent region[FCO_MAX_REGIONS])
{
	struct fellow_cache_seglist *fcsl;
	struct fellow_disk_seglist *fdsl;
	unsigned n;

	CHECK_OBJ_NOTNULL(fco, FELLOW_CACHE_OBJ_MAGIC);

	n = fellow_seglist_regions(ffd, &fco->seglist, region, 0);
	assert(n <= FCO_MAX_REGIONS);
	DBG("seglist_regions %u", n);
#define FDO_AUXATTR(U, l)						\
	if (fco->aa_##l##_seg.state != FCS_USABLE)			\
		fellow_seg_regions(ffd, &fco->aa_##l##_seg, region, &n);
#include "tbl/fellow_obj_attr.h"
	DBG("+auxattr %u", n);
	assert(n <= FCO_MAX_REGIONS);

	fcsl = &fco->seglist;
	fdsl = fcsl->fdsl;

	while ((fcsl = fcsl->next) != NULL) {
		/* fcsl is ahead of fdsl by one (fcsl is next of fdsl) */
		CHECK_OBJ(fcsl, FELLOW_CACHE_SEGLIST_MAGIC);
		CHECK_OBJ_NOTNULL(fdsl, FELLOW_DISK_SEGLIST_MAGIC);

		region[n++] = fdsl->next;
		assert(n <= FCO_MAX_REGIONS);
		fdsl = fcsl->fdsl;
	}
	DBG("+seglists %u", n);
	AZ(fdsl->next.off);
	return (n);
}

/* fco mtx must be held, will be unlocked */
static void
fellow_cache_obj_free(struct fellow_cache *fc, struct fellow_cache_obj **fcop)
{
	struct fellow_cache_obj *fco;
	struct fellow_cache_seg *fcs;

	TAKE_OBJ_NOTNULL(fco, fcop, FELLOW_CACHE_OBJ_MAGIC);

	fcs = FCO_FCS(fco);
	AZ(FCO_REFCNT(fco));
	assert_cache_seg_consistency(fcs);
	fellow_cache_seg_transition_locked(fc, FCO_FCS(fco),
	    FCO_STATE(fco), FCO_EVICT);

	DBG("fco %p", fco);
	fellow_cache_seglist_free(fc, &fco->seglist);

#define FDO_AUXATTR(U, l)				\
	fellow_cache_seg_auxattr_free(fc, &fco->aa_##l##_seg);
#include "tbl/fellow_obj_attr.h"

	AZ(fcs->fco_infdb);
	AZ(fcs->fcs_onlru);

	//lint --e{455} mutex unlocked
	AZ(pthread_mutex_unlock(&fco->mtx));
	AZ(pthread_mutex_destroy(&fco->mtx));
	AZ(pthread_cond_destroy(&fco->cond));

	fellow_cache_seg_free(fc, fcs, 0);

	if (fco->fco_mem.bits) {
		assert(fco->fco_mem.ptr == fco);
		buddy_return1_ptr_page(fc->membuddy, &fco->fco_mem);
	}
}

/*
 * evict from stvfe_mutate():
 *
 * state must be FCO_INCORE, object is on LRU, must be referenced
 *
 * lru and fco mtx are held
 *
 * similar in structure to fellow_cache_obj_deref(), but
 * - keep reference
 * - remove exp
 */

void
fellow_cache_obj_evict_mutate(struct fellow_cache *fc,
    struct fellow_cache_obj *fco)
{
	struct fellow_cache_seg *fcs;

	CHECK_OBJ_NOTNULL(fc, FELLOW_CACHE_MAGIC);
	CHECK_OBJ_NOTNULL(fco, FELLOW_CACHE_OBJ_MAGIC);

	CHECK_OBJ_NOTNULL(fco, FELLOW_CACHE_OBJ_MAGIC);
	fcs = FCO_FCS(fco);

	assert_cache_seg_consistency(fcs);
	assert(fco->logstate == FCOL_INLOG);

	// fdb
	AN(fcs->fco_infdb);
	fcs->fco_infdb = 0;
	AN(FCO_REFCNT(fco));
	AZ(pthread_mutex_lock(&fc->fdb_mtx));
	(void) VRBT_REMOVE(fellow_cache_fdb_head, &fc->fdb_head, fco);
	AZ(pthread_mutex_unlock(&fc->fdb_mtx));

	// state, oc (relevant for lru)
	assert(FCO_STATE(fco) == FCO_INCORE);
	FCO_STATE(fco) = FCO_EVICT;
	AN(fco->oc);
	fco->oc = NULL;

	// lru
	VTAILQ_REMOVE(&fc->lru_head, fcs, lru_list);
	AN(fcs->fcs_onlru);
	fcs->fcs_onlru = 0;

	assert_cache_seg_consistency(fcs);
}

static inline void
fellow_disk_seg_init(struct fellow_disk_seg *fds, uint8_t fht)
{

	assert(fht > FH_NONE);
	assert(fht < FH_LIM);
	INIT_OBJ(fds, FELLOW_DISK_SEG_MAGIC);
	fds->fht = fht;
}

static void
fellow_disk_seglist_init(struct fellow_disk_seglist *fdsl,
    uint16_t ldsegs, uint8_t fht)
{
	unsigned u;

	INIT_OBJ(fdsl, FELLOW_DISK_SEGLIST_MAGIC);
	fdsl->version = 1;
	fdsl->lsegs = ldsegs;
	fdsl->fht = fht;
	for (u = 0; u < ldsegs; u++)
		fellow_disk_seg_init(&fdsl->segs[u], fht);
}

static inline void
fellow_cache_seg_init(struct fellow_cache_seg *fcs,
    struct fellow_cache_obj *fco, enum fcos_state state)
{

	CHECK_OBJ_NOTNULL(fco, FELLOW_CACHE_OBJ_MAGIC);
	INIT_OBJ(fcs, FELLOW_CACHE_SEG_MAGIC);
	fcs->state = state;
	fcs->fco = fco;
}

static inline void
fellow_cache_seg_associate(struct fellow_cache_seg *fcs,
    struct fellow_disk_seg *fds, enum fcos_state state)
{

	assert(fcs->state == FCS_INIT);
	assert_cache_seg_consistency(fcs);
	fcs->disk_seg = fds;
	fellow_cache_seg_transition_locked_notincore(fcs, state);
}

/*
 * cache seglist must be able to hold all disk seglist
 */
static void
fellow_cache_seglist_associate(
    struct fellow_cache_seglist *fcsl,
    struct fellow_disk_seglist *fdsl, enum fcos_state state)
{
	struct fellow_cache_seg *fcs;
	struct fellow_disk_seg *fds;
	unsigned u;

	CHECK_OBJ_NOTNULL(fcsl, FELLOW_CACHE_SEGLIST_MAGIC);
	CHECK_OBJ_NOTNULL(fdsl, FELLOW_DISK_SEGLIST_MAGIC);

	assert(fcsl->lsegs >= fdsl->lsegs);
	AZ(fcsl->fdsl);
	fcsl->fdsl = fdsl;

	for (u = 0; u < fdsl->nsegs; u++) {
		fcs = &fcsl->segs[u];
		fds = &fdsl->segs[u];

		fellow_cache_seg_associate(fcs, fds, state);
	}
	for (; u < fdsl->lsegs; u++) {
		fcs = &fcsl->segs[u];
		fds = &fdsl->segs[u];

		fellow_cache_seg_associate(fcs, fds, FCS_USABLE);
	}
}

static void
fellow_cache_seglist_init(struct fellow_cache_seglist *fcsl,
    size_t space, struct fellow_cache_obj *fco)
{
	uint16_t u;
	size_t lsegs;

	AN(fcsl);
	assert(space > sizeof *fcsl);
	INIT_OBJ(fcsl, FELLOW_CACHE_SEGLIST_MAGIC);
	lsegs = space / sizeof *fcsl->segs;
	assert(lsegs <= UINT16_MAX);
	fcsl->lsegs = (uint16_t)lsegs;
	for (u = 0; u < fcsl->lsegs; u++)
		fellow_cache_seg_init(&fcsl->segs[u], fco, FCS_INIT);
}

static void
fellow_disk_seglist_fini(struct fellow_disk_seglist *fdsl)
{
	unsigned u;

	CHECK_OBJ_NOTNULL(fdsl, FELLOW_DISK_SEGLIST_MAGIC);
	assert(fdsl->version == 1);
	// AN(fdsl->nsegs); // not possible for the embedded seglist
	assert(fdsl->nsegs <= fdsl->lsegs);
	// no zero segments active
	for (u = 0; u < fdsl->nsegs; u++) {
		AN(fdsl->segs[u].seg.off);
		AN(fdsl->segs[u].seg.size);
	}
	fh(fdsl->fht, fdsl->fh,
	    (char *)fdsl + FELLOW_DISK_SEGLIST_CHK_START,
	    FELLOW_DISK_SEGLIST_CHK_LEN(fdsl));
}

/*
 * region is from the previous disk seglist
 *
 * the first seglist is embedded
 */
static void
fellow_cache_seglists_write(struct fellow_busy *fbo,
    const struct fellow_cache_seglist *fcsl,
    struct buddy_off_extent reg)
{
	struct fellow_disk_seglist *fdsl;
	struct fellow_cache_obj *fco;
	struct fellow_busy_io *fbio, fbiostk[1];

	CHECK_OBJ_NOTNULL(fbo, FELLOW_BUSY_MAGIC);

	fco = fbo->fco;
	CHECK_OBJ_NOTNULL(fco, FELLOW_CACHE_OBJ_MAGIC);

	while (fcsl) {
		CHECK_OBJ_NOTNULL(fcsl, FELLOW_CACHE_SEGLIST_MAGIC);
		AN(reg.off);
		AN(reg.size);
		assert(reg.size <= FIO_MAX);
		fdsl = fcsl->fdsl;

		fellow_disk_seglist_fini(fdsl);

		AZ(pthread_mutex_lock(&fco->mtx));
		fbo->io_outstanding++;
		fbio = fellow_busy_io_get(fbo, fbiostk);
		AZ(pthread_mutex_unlock(&fco->mtx));

		AN(fbio);
		fbio->type = FBIO_SEGLIST;
		fbio->u.seglist.reg = reg;
		fbio->u.seglist.fdsl = fdsl;

		reg = fdsl->next;
		fcsl = fcsl->next;

		fellow_busy_io_submit(fbio);
	}
}

#define fdslwrong(x) return(FC_ERRSTR("disk seglist wrong " x));

static const char *
fellow_disk_seglist_check(const struct fellow_disk_seglist *fdsl)
{

	AN(fdsl);
	if (FC_INJ || fdsl->magic != FELLOW_DISK_SEGLIST_MAGIC)
		fdslwrong("magic");
	if (fdsl->version != 1)
		fdslwrong("version");
	if (fdsl->nsegs > fdsl->lsegs)
		fdslwrong("nsegs");
	if (fdsl->fht >= FH_LIM)
		fdslwrong("hash type");
	if (FC_INJ || fhcmp(fdsl->fht, fdsl->fh,
		(char *)fdsl + FELLOW_DISK_SEGLIST_CHK_START,
		FELLOW_DISK_SEGLIST_CHK_LEN(fdsl)))
		fdslwrong("chksum");
	return (NULL);
}

/*
 * starts with the embedded disk/cache seglist
 */

static const char *
fellow_cache_seglists_load(const struct fellow_cache *fc,
    struct fellow_cache_obj *fco,
    struct fellow_cache_seglist *fcsl, struct fellow_disk_seglist *fdsl)
{
	struct buddy_ptr_extent fdsl_mem, fcsl_mem;
	struct fellow_cache_seglist *ofcsl;
	struct buddy_off_extent next;
	const char *err;
	ssize_t ssz;
	size_t sz;

	CHECK_OBJ_NOTNULL(fc, FELLOW_CACHE_MAGIC);

	err = fellow_disk_seglist_check(fdsl);
	if (err != NULL)
		return (err);

	while (1) {
		CHECK_OBJ_NOTNULL(fcsl, FELLOW_CACHE_SEGLIST_MAGIC);
		CHECK_OBJ_NOTNULL(fdsl, FELLOW_DISK_SEGLIST_MAGIC);

		fellow_cache_seglist_associate(fcsl, fdsl, FCS_DISK);

		next = fdsl->next;
		if (next.off == 0 || next.size == 0) {
			AZ(next.off);
			AZ(next.size);
			break;
		}

		fcsl_mem = buddy_ptr_extent_nil;
		fdsl_mem = buddy_alloc1_ptr_extent_wait(fc->membuddy, FEP_META,
		    next.size, 0);
		fdsl = fdsl_mem.ptr;
		if (FC_INJ || fdsl == NULL) {
			err = FC_ERRSTR("disk seglist fdsl alloc failed");
			goto err;
		}
		assert(fdsl_mem.size >= next.size);

		ssz = fellow_io_pread_sync(fc->ffd, fdsl, next.size, next.off);
		if (FC_INJ || ssz < 0 || (size_t)ssz != next.size) {
			err = FC_ERRSTR("disk seglist read failed");
			goto err;
		}
		err = fellow_disk_seglist_check(fdsl);
		if (err != NULL)
			goto err;

		ofcsl = fcsl;

		sz = fellow_cache_seglist_size(fcsl, fdsl->lsegs);

		fcsl_mem = buddy_alloc1_ptr_extent_wait(fc->membuddy, FEP_META,
		    sz, 0);
		fcsl = fcsl_mem.ptr;
		if (FC_INJ || fcsl == NULL) {
			err = FC_ERRSTR("disk seglist fcsl alloc failed");
			goto err;
		}

		fellow_cache_seglist_init(fcsl,
		    fcsl_mem.size - sizeof *fcsl, fco);
		fcsl->fdsl_sz = fdsl_mem.size;
		fcsl->fcsl_sz = fcsl_mem.size;

		AZ(ofcsl->next);
		ofcsl->next = fcsl;
	}

	return (NULL);
  err:
	if (fdsl_mem.ptr != NULL)
		buddy_return1_ptr_extent(fc->membuddy, &fdsl_mem);
	if (fcsl_mem.ptr != NULL)
		buddy_return1_ptr_extent(fc->membuddy, &fcsl_mem);
	return (err);
}

/*
 * MEM LAYOUT
 *
 *     fellow_disk_obj
 * ? x va_data
 *     fellow_disk_seglist
 * ? x fellow_disk_seg
 * dsk_sz = sum(^^)
 *
 * mem_sz = sum(vv)
 *     fellow_cache_obj
 *     [fellow_cache_seglist]	embedded in cache_obj
 * n x fellow_cache_seg
 *
 * for a call from fellow_busy_obj_alloc, we know how many disk_seg
 * for a call from fellow_cache_obj_prepread, we will guess (0 argument)
 *
 * we make one or two allocations, depending on whether the
 * second part fits in the same LOG2UP or not
 *
 * XXX
 * - use two allocations always?
 * - use struct reqs?
 * - collapse allocation with fellow_budy_obj_alloc?
 */

static struct fellow_cache_res
fellow_cache_obj_new(const struct fellow_cache *fc,
    size_t dsk_sz, unsigned nseg_guess)
{
	struct fellow_disk_obj *fdo;
	struct fellow_cache_obj *fco;
	struct fellow_cache_res fcr;
	struct fellow_cache_seg *fcs;
	struct buddy_ptr_page fco_mem = buddy_ptr_page_nil;
	struct buddy_ptr_extent fdo_mem = buddy_ptr_extent_nil;
	size_t asz, mem_sz, ssz;
	uint8_t bits;
	unsigned u;

	DBG("arg dsk_sz %zu nseg_guess %u", dsk_sz, nseg_guess);

	if (nseg_guess == 0) {
		asz = sizeof(struct fellow_disk_obj) +
		    sizeof(struct fellow_disk_seglist);
		assert(dsk_sz > asz);
		ssz = (dsk_sz - asz) / sizeof(struct fellow_disk_seg);
		assert(ssz < UINT_MAX);
		nseg_guess = (unsigned)ssz;
		if (nseg_guess == 0)
			nseg_guess = 1;
	}

	mem_sz = sizeof *fco + nseg_guess * sizeof *fcs;

	asz = (size_t)1 << log2up(dsk_sz);

//	DBG("dsk_sz %zu nseg_guess %u mem_sz %zu asz %zu",
//	    dsk_sz, nseg_guess, mem_sz, asz);

	fco = NULL;
	if (dsk_sz + mem_sz > asz) {
		// fco etc in extra allocation
		u = log2up(mem_sz);
		assert(u <= UINT8_MAX);
		bits = (uint8_t)u;
		fco_mem = buddy_alloc1_ptr_page_wait(fc->membuddy,
		    FEP_NORMAL, bits, 0);
		if (FC_INJ || fco_mem.ptr == NULL) {
			fcr = FCR_ALLOCFAIL("fco_mem alloc failed");
			goto err;
		}
		fco = fco_mem.ptr;
		asz = dsk_sz;
	}

	fdo_mem = buddy_alloc1_ptr_extent_wait(fc->membuddy,
	    FEP_NORMAL, asz, 0);

	if (FC_INJ || fdo_mem.ptr == NULL) {
		fcr = FCR_ALLOCFAIL("fdo_mem alloc failed");
		goto err;
	}

	fdo = fdo_mem.ptr;

	if (fco == NULL)
		fco = (void *)((unsigned char *)fdo + dsk_sz);

	INIT_OBJ(fco, FELLOW_CACHE_OBJ_MAGIC);
	DBG("fco %p", fco);
	fco->fco_mem = fco_mem;

	AZ(pthread_mutex_init(&fco->mtx, NULL));
	AZ(pthread_cond_init(&fco->cond, NULL));

	if (fco_mem.bits) {
		ssz = ((size_t)1 << fco_mem.bits) - sizeof *fco;
	}
	else {
		ssz = fdo_mem.size;
		assert(ssz > dsk_sz);
		ssz -= dsk_sz;
		assert(ssz > sizeof *fco);
		ssz -= sizeof *fco;
	}

	fellow_cache_seglist_init(&fco->seglist, ssz, fco);

#define FDO_AUXATTR(U, l)						\
	fellow_cache_seg_init(&fco->aa_##l##_seg, fco, FCS_INIT);
#include "tbl/fellow_obj_attr.h"

	fcs = &fco->fdo_fcs;
	fellow_cache_seg_init(fcs, fco, FCO_INIT);

	fcs->alloc = fdo_mem;

	return (FCR_OK(fco));
  err:
	if (fco_mem.ptr != NULL)
		buddy_return1_ptr_page(fc->membuddy, &fco_mem);
	if (fdo_mem.ptr != NULL)
		buddy_return1_ptr_extent(fc->membuddy, &fdo_mem);

	fellow_cache_res_check(fc, fcr);
	return (fcr);
}

struct fellow_cache_res
fellow_busy_obj_alloc(struct fellow_cache *fc,
    struct fellow_cache_obj **fcop, uintptr_t *priv2,
    unsigned wsl)
{
	struct fellow_disk_obj *fdo;
	struct fellow_disk_seg *fds;
	struct fellow_disk_seglist *fdsl;
	struct fellow_cache_obj *fco;
	struct fellow_cache_seg *fcs;
	struct fellow_cache_res fcr;
	struct fellow_busy *fbo;
	struct buddy_ptr_extent fbo_mem;
	struct buddy_off_extent fds_seg;
	size_t sz, asz, dsk_sz;
	uint16_t ldsegs;

	fbo_mem = buddy_alloc1_ptr_extent_wait(fc->membuddy, FEP_NORMAL,
	    sizeof *fbo, 0);
	fbo = fbo_mem.ptr;
	if (fbo == NULL) {
		fcr = FCR_ALLOCFAIL("fbo_mem");
		fellow_cache_res_check(fc, fcr);
		return (fcr);
	}
	INIT_OBJ(fbo, FELLOW_BUSY_MAGIC);
	fbo->fbo_mem = fbo_mem;

	// XXX peek for known content-length=

	// round up to disk block such that there
	// is space for at least one disk_seg
	wsl = (unsigned)PRNDUP(wsl);
	asz = sizeof(struct fellow_disk_obj) +
	    wsl +
	    sizeof(struct fellow_disk_seglist) +
	    sizeof(struct fellow_disk_seg);

	dsk_sz = fellow_rndup(fc->ffd, asz);
	assert(PAOK(dsk_sz));

//	DBG("roundup fdo + fdsl dsk_sz %zu", dsk_sz);

	assert(dsk_sz > asz);
	sz = ((dsk_sz - asz) + sizeof(struct fellow_disk_seg))
	    / sizeof(struct fellow_disk_seg);
	assert(sz <= UINT16_MAX);
	ldsegs = (uint16_t)sz;

	// uring has signed 32bit length
	assert(dsk_sz <= FIO_MAX);

	// XXX delay to allow new objects during FP_INIT ?
	fds_seg = FC_INJ ? buddy_off_extent_nil :
	    fellow_region_alloc(fc->ffd, FEP_NORMAL, dsk_sz, 0);
	if (fds_seg.off < 0) {
		fcr = FCR_ALLOCFAIL("fds region");
		fellow_cache_res_check(fc, fcr);
		buddy_return1_ptr_extent(fc->membuddy, &fbo->fbo_mem);
		return (fcr);
	}

	fcr = fellow_cache_obj_new(fc, dsk_sz, ldsegs);
	if (fcr.status != fcr_ok) {
		fellow_cache_res_check(fc, fcr);
		fellow_region_free(fc->ffd, &fds_seg);
		buddy_return1_ptr_extent(fc->membuddy, &fbo->fbo_mem);
		return (fcr);
	}
	fco = fcr.r.ptr;

	CHECK_OBJ(fco, FELLOW_CACHE_OBJ_MAGIC);
	fcs = &fco->fdo_fcs;
	CHECK_OBJ(fcs, FELLOW_CACHE_SEG_MAGIC);

	// disk object in memory
	fdo = fcs->alloc.ptr;
	memset(fdo, 0, fcs->alloc.size);
	fdo->magic = FELLOW_DISK_OBJ_MAGIC;
	fdo->version = 1;
	fdo->va_data_len = wsl;

	// disk obj init (allocation happened before)
	fds = &fdo->fdo_fds;
	fellow_disk_seg_init(fds, fc->tune->hash_obj);

	fds->seg = fds_seg;

	fcs->disk_seg = fds;
	fco->fdb = region_fdb(&fds->seg);
	AN(fcop);
	*fcop = fco;
	AN(priv2);
	*priv2 = fco->fdb.fdb;

	fbo->va_data = fdo->va_data;
#define FDO_AUXATTR(U, l)						\
	fellow_disk_seg_init(&fdo->aa_##l##_seg, fc->tune->hash_obj);	\
	fellow_cache_seg_associate(&fco->aa_##l##_seg,			\
	    &fdo->aa_##l##_seg, FCS_USABLE);
#include "tbl/fellow_obj_attr.h"

	// we prepare the state the object will have
	// when inserted -- 2nd ref is for write fellow_cache_async_complete()
	fcs->refcnt = 2;
	fellow_cache_seg_transition_locked_notincore(fcs, FCO_BUSY);

	fdsl = (void *)((uint8_t *)fdo + sizeof *fdo + wsl);
	assert(PAOK(fdsl));

	fellow_disk_seglist_init(fdsl, ldsegs, fc->tune->hash_obj);
	fellow_cache_seglist_associate(&fco->seglist, fdsl, FCS_USABLE);
//	DBG("fdsl lsegs %u fcsl lsegs %u", fdsl->lsegs, fco->seglist.lsegs);

	fbo->fc = fc;
	fbo->fco = fco;
	fbo->body_seglist = &fco->seglist;
	AZ(fbo->body_seg);

	AN(fbo->body_seglist);
	AN(fbo->body_seglist->lsegs);
	AZ(fbo->body_seglist->fdsl->nsegs);

	/* keep one io outstanding until unbusy is called
	 * to prevent the object from completing to FCO_INCORE
	 *
	 * fellow_cache_obj_unbusy does not take an IO, it owns this one
	 */
	fbo->io_outstanding = 1;

	return (FCR_OK(fbo));
}

/* allocate a new region for body, aux attr or seglist
 *
 * aux attr use: We do not yet coalesce allocations because
 * there only is a single aux attr, so the one segment gets
 * its own region
 */
static struct buddy_off_extent *
fellow_busy_region_alloc(struct fellow_busy *fbo, size_t size, int8_t cram)
{
	struct buddy_off_extent *fdr;
	struct fellow_cache *fc;

	CHECK_OBJ_NOTNULL(fbo, FELLOW_BUSY_MAGIC);

	AN(size);
	if (size > FIO_MAX)
		size = FIO_MAX;

	fc = fbo->fc;
	CHECK_OBJ_NOTNULL(fc, FELLOW_CACHE_MAGIC);

	/* stop cramming if we used half the regions already */
	if (fbo->nregion < FCO_MAX_REGIONS / 2)
		cram = 0;

	assert(fbo->nregion <= FCO_MAX_REGIONS);
	fdr = &fbo->region[fbo->nregion];

	*fdr = FC_INJ ? buddy_off_extent_nil :
	    fellow_region_alloc(fc->ffd, FEP_NORMAL, size, cram);

	if (fdr->off < 0) {
		return (NULL);
	}

	AN(fdr->size);
	AZ(fdr->size & FELLOW_BLOCK_ALIGN);

	fbo->nregion++;

	return (fdr);
}

/*
 * return an unused region for seglist trim
 */
static void
fellow_busy_region_free(struct fellow_busy *fbo, struct buddy_off_extent *fdr)
{
	struct fellow_cache *fc;
	unsigned u;

	CHECK_OBJ_NOTNULL(fbo, FELLOW_BUSY_MAGIC);

	fc = fbo->fc;
	CHECK_OBJ_NOTNULL(fc, FELLOW_CACHE_MAGIC);

	AN(fbo->nregion);
	u = fbo->nregion;
	while (u--) {
		if (fdr != &fbo->region[u] &&
		    fdr->off != fbo->region[u].off &&
		    fdr->size != fbo->region[u].size)
			continue;
		fellow_region_free(fc->ffd, fdr);
		fbo->nregion--;
		assert(u <= fbo->nregion);
		if (u < fbo->nregion) {
			memmove(&fbo->region[u], &fbo->region[u] + 1,
			    (fbo->nregion - u) * sizeof *fdr);
		}
		return;
	}
	WRONG("region to free not found");
}

static struct fellow_cache_res
fellow_body_seglist_alloc(struct fellow_busy *fbo,
    struct fellow_cache_seglist *ofcsl)
{
	struct fellow_disk_seglist *ofdsl, *fdsl;
	struct fellow_cache_seglist *fcsl;
	struct fellow_cache *fc;
	struct buddy_off_extent *fdr;
	struct buddy_ptr_extent fdsl_mem, fcsl_mem;
	struct buddy_reqs *reqs;
	unsigned chunk_exponent, u;
	uint16_t ldsegs;
	size_t sz;

	CHECK_OBJ_NOTNULL(fbo, FELLOW_BUSY_MAGIC);

	CHECK_OBJ_NOTNULL(ofcsl, FELLOW_CACHE_SEGLIST_MAGIC);
	AZ(ofcsl->next);

	ofdsl = ofcsl->fdsl;
	AZ(ofdsl->next.off);
	AZ(ofdsl->next.size);

	fc = fbo->fc;
	CHECK_OBJ_NOTNULL(fc, FELLOW_CACHE_MAGIC);
	CHECK_OBJ_NOTNULL(fc->tune, STVFE_TUNE_MAGIC);
	chunk_exponent = fc->tune->chunk_exponent;


	assert(fbo->sz_estimate > fbo->sz_returned);

	sz = fbo->sz_estimate - fbo->sz_returned;
	sz >>= chunk_exponent;
	assert(sz <= UINT16_MAX);
	ldsegs = (uint16_t)sz;
	ldsegs++;

	sz = fellow_disk_seglist_size(fdsl, ldsegs);

	fdr = fellow_busy_region_alloc(fbo, sz, 0);
	if (fdr == NULL)
		return (FCR_ALLOCFAIL("seglist disk region"));
	ofdsl->next = *fdr;

	sz = fdr->size;
	assert(sz > sizeof *fdsl);
	sz -= sizeof *fdsl;
	sz /= sizeof *fdsl->segs;
	assert(sz <= UINT16_MAX);
	ldsegs = (uint16_t)sz;

	reqs = BUDDY_REQS_STK(fbo->fc->membuddy, 2);
	BUDDY_REQS_PRI(reqs, FEP_META);
	AN(buddy_req_extent(reqs, fdr->size, 0));
	AN(buddy_req_extent(reqs, fellow_cache_seglist_size(fcsl, ldsegs), 0));
	u = buddy_alloc_wait(reqs);
	if (FC_INJ || u != 2) {
		buddy_alloc_wait_done(reqs);
		fellow_busy_region_free(fbo, fdr);
		return (FCR_ALLOCFAIL("seglist memory"));
	}

	fdsl_mem = buddy_get_ptr_extent(reqs, 0);
	fcsl_mem = buddy_get_ptr_extent(reqs, 1);

	buddy_alloc_wait_done(reqs);

	fdsl = fdsl_mem.ptr;
	AN(fdsl);
	assert(fdsl_mem.size == fdr->size);
	fellow_disk_seglist_init(fdsl, ldsegs, fc->tune->hash_obj);

	fcsl = fcsl_mem.ptr;
	AN(fcsl);
	fellow_cache_seglist_init(fcsl, fcsl_mem.size - sizeof *fcsl, fbo->fco);

	fellow_cache_seglist_associate(fcsl, fdsl, FCS_USABLE);

	fcsl->fcsl_sz = fcsl_mem.size;
	fcsl->fdsl_sz = fdsl_mem.size;
	ofcsl->next = fcsl;

	return (FCR_OK(fcsl));
}

/* allocate a new region for body data */
static size_t
fellow_busy_body_seg_alloc(struct fellow_busy *fbo,
    struct fellow_disk_seg *fds, size_t want)
{
	struct fellow_body_region *fbr;
	struct buddy_off_extent *fdr;
	struct fellow_cache *fc;
	size_t spc;

	CHECK_OBJ_NOTNULL(fbo, FELLOW_BUSY_MAGIC);
	CHECK_OBJ_NOTNULL(fds, FELLOW_DISK_SEG_MAGIC);
	fc = fbo->fc;
	CHECK_OBJ_NOTNULL(fc, FELLOW_CACHE_MAGIC);
	CHECK_OBJ_NOTNULL(fc->tune, STVFE_TUNE_MAGIC);

	fbr = &fbo->body_region;
	AZ(fds->seg.off);
	AZ(fds->seg.size);

	want = fellow_rndup(fc->ffd,  want);
	AZ(want & FELLOW_BLOCK_ALIGN);

	fdr = fbr->reg;
	if (fdr == NULL || fbr->len == fdr->size) {

		memset(fbr, 0, sizeof *fbr);

		assert(fbo->sz_estimate > fbo->sz_dskalloc);
		spc = fbo->sz_estimate - fbo->sz_dskalloc;

		fdr = fellow_busy_region_alloc(fbo, spc, fc->tune->cram);
		if (fdr == NULL)
			return (0);

		fbr->reg = fdr;
		AZ(fbr->len);
		fbo->sz_dskalloc += fdr->size;
	}

	assert(fbr->len < fdr->size);

	spc = fdr->size - fbr->len;
	AZ(spc & FELLOW_BLOCK_ALIGN);

	if (spc < want)
		want = spc;

	fds->seg.off = fdr->off + (buddyoff_t)fbr->len;
	fds->seg.size = want;
	fds->segnum = fbr->segnum++;

	fbr->len += want;

	return (want);
}

/* return an fds to the body seg (for failed memalloc) */
static void
fellow_busy_body_seg_return(struct fellow_busy *fbo,
    struct fellow_disk_seg *fds)
{
	struct fellow_body_region *fbr;
	struct buddy_off_extent *fdr;

	CHECK_OBJ_NOTNULL(fbo, FELLOW_BUSY_MAGIC);

	AN(fds->seg.off);
	AN(fds->seg.size);

	fbr = &fbo->body_region;
	fdr = fbr->reg;
	AN(fdr);

	assert(fbr->len >= fds->seg.size);
	fbr->len -= fds->seg.size;
	assert(fds->seg.off == fdr->off + (buddyoff_t)fbr->len);
	fbr->segnum--;

	fds->seg.off = 0;
	fds->seg.size = 0;
}


static void
fellow_cache_seg_fini(const struct fellow_cache_seg *fcs)
{
	struct fellow_disk_seg	*fds;

	CHECK_OBJ_NOTNULL(fcs, FELLOW_CACHE_SEG_MAGIC);
	fds = fcs->disk_seg;
	CHECK_OBJ_NOTNULL(fds, FELLOW_DISK_SEG_MAGIC);

	AN(fcs->alloc.ptr);
	AN(fcs->len);
	assert(fcs->len <= fds->seg.size);
	fds->seg.size = fcs->len;
	fh(fds->fht, fds->fh, fcs->alloc.ptr, fcs->len);
}

static inline void
fellow_cache_seg_transition_locked_notincore(struct fellow_cache_seg *fcs,
    enum fcos_state to)
{

	CHECK_OBJ_NOTNULL(fcs, FELLOW_CACHE_SEG_MAGIC);

	/*
	 * to/from INCORE requires lru operations
	 * see fellow_cache_seg_transition*
	 */
	assert(! FCOS_IS(fcs->state, INCORE));
	assert(! FCOS_IS(to, INCORE));

	assert(fcs->state != to);
	assert_fcos_transition(fcs->state, to);
	fcs->state = to;
	assert_cache_seg_consistency(fcs);
}

static inline void
fellow_cache_seg_transition_locked(
    struct fellow_cache *fc, struct fellow_cache_seg *fcs,
    enum fcos_state from, enum fcos_state to)
{
	int o, n;

	DBG("%p %s %s", fcs, fcos_state_s[from], fcos_state_s[to]);
	o = fellow_cache_shouldlru(from, fcs->fco->oc, fcs->refcnt);
	n = fellow_cache_shouldlru(to, fcs->fco->oc, fcs->refcnt);

	assert(fcs->state == from);
	fcs->state = to;

	fellow_cache_lru_chg(fc, fcs, n - o);
}

static void
fellow_cache_seg_transition(
    struct fellow_cache *fc,
    struct fellow_cache_seg *fcs,
    enum fcos_state from, enum fcos_state to)
{
	struct fellow_cache_obj *fco;

	CHECK_OBJ_NOTNULL(fcs, FELLOW_CACHE_SEG_MAGIC);
	fco = fcs->fco;
	CHECK_OBJ_NOTNULL(fco, FELLOW_CACHE_OBJ_MAGIC);
	assert(from != to);

	assert_fcos_transition(from, to);

	AZ(pthread_mutex_lock(&fco->mtx));
	fellow_cache_seg_transition_locked(fc, fcs, from, to);
	assert_cache_seg_consistency(fcs);
	if (fcs->refcnt)
		AZ(pthread_cond_broadcast(&fco->cond));
	AZ(pthread_mutex_unlock(&fco->mtx));
}

/* return old refcnt */
static inline unsigned
fellow_cache_seg_ref_locked(struct fellow_cache *fc,
    struct fellow_cache_seg *fcs)
{
	unsigned refcnt;
	int o, n;

	refcnt = fcs->refcnt++;

	o = fellow_cache_shouldlru(fcs->state, fcs->fco->oc, refcnt);
	n = fellow_cache_shouldlru(fcs->state, fcs->fco->oc, fcs->refcnt);

	fellow_cache_lru_chg(fc, fcs, n - o);
	return (refcnt);
}

/* return new refcnt */
static inline unsigned
fellow_cache_seg_deref_locked(struct fellow_cache *fc,
    struct fellow_cache_seg *fcs)
{
	unsigned refcnt;
	int o, n;

	AN(fcs->refcnt);
	refcnt = --fcs->refcnt;

	o = fellow_cache_shouldlru(fcs->state, fcs->fco->oc, refcnt + 1);
	n = fellow_cache_shouldlru(fcs->state, fcs->fco->oc, refcnt);

	fellow_cache_lru_chg(fc, fcs, n - o);
	return (refcnt);
}

static void
fellow_cache_seg_unbusy(struct fellow_busy *fbo, struct fellow_cache_seg *fcs){
	const enum fcos_state from = FCS_BUSY, to = FCS_WRITING;
	struct fellow_cache_obj *fco;
	struct fellow_busy_io *fbio, fbiostk[1];

	CHECK_OBJ_NOTNULL(fbo, FELLOW_BUSY_MAGIC);
	CHECK_OBJ_NOTNULL(fbo->fc, FELLOW_CACHE_MAGIC);
	CHECK_OBJ_NOTNULL(fcs, FELLOW_CACHE_SEG_MAGIC);
	fco = fcs->fco;
	CHECK_OBJ_NOTNULL(fco, FELLOW_CACHE_OBJ_MAGIC);
	assert(fco == fbo->fco);
	AN(fcs->refcnt);
	AN(fcs->len);
	assert_fcos_transition(from, to);

	AZ(pthread_mutex_lock(&fco->mtx));
	fbo->io_outstanding++;
	fbio = fellow_busy_io_get(fbo, fbiostk);
	fellow_cache_seg_transition_locked(fbo->fc, fcs, from, to);
	assert_cache_seg_consistency(fcs);
#ifdef WAIT_FOR_FCS_WRITING
	// we should never have a wait for FCS_WRITING
	if (fcs->refcnt)
		AZ(pthread_cond_broadcast(&fco->cond));
#endif
	AZ(pthread_mutex_unlock(&fco->mtx));

	fellow_cache_seg_fini(fcs);

	AN(fbio);
	fbio->type = FBIO_SEG;
	fbio->u.fcs = fcs;

	fellow_busy_io_submit(fbio);
}

/* XXX can save some io for inlog == 0 */

static void
fellow_cache_obj_unbusy(struct fellow_busy *fbo, enum fcol_state wantlog)
{
	const enum fcos_state from = FCO_BUSY, to = FCO_WRITING;
	struct fellow_cache_obj *fco, *ofco;
	struct fellow_cache_seg *fcs;
	struct fellow_cache *fc;
	struct fellow_busy_io *fbio, fbiostk[1];

	CHECK_OBJ_NOTNULL(fbo, FELLOW_BUSY_MAGIC);
	fc = fbo->fc;
	CHECK_OBJ_NOTNULL(fc, FELLOW_CACHE_MAGIC);
	fco = fbo->fco;
	CHECK_OBJ_NOTNULL(fco, FELLOW_CACHE_OBJ_MAGIC);

	fcs = FCO_FCS(fco);
	assert_fcos_transition(from, to);

	assert(wantlog == FCOL_WANTLOG || wantlog == FCOL_NOLOG);

	AZ(pthread_mutex_lock(&fco->mtx));
	AZ(fcs->fco_infdb);
	fcs->fco_infdb = 1;
	AZ(pthread_mutex_lock(&fc->fdb_mtx));
	ofco = VRBT_INSERT(fellow_cache_fdb_head, &fc->fdb_head, fco);
	AZ(ofco);
	AZ(pthread_mutex_unlock(&fc->fdb_mtx));

	if (likely(fco->logstate == FCOL_DUNNO))
		fco->logstate = wantlog;
	else
		assert(fco->logstate == FCOL_TOOLATE);

	/*
	 * fco has its io_outstanding from fellow_busy_obj_alloc()
	 * fbo->io_outstanding++;
	 */
	fbio = fellow_busy_io_get(fbo, fbiostk);
	fellow_cache_seg_transition_locked(fc, fcs, from, to);
	assert_cache_seg_consistency(fcs);
#ifdef WAIT_FOR_FCS_WRITING
	// we should never have a wait for FCS_WRITING
	if (fcs->refcnt)
		AZ(pthread_cond_broadcast(&fco->cond));
#endif
	AZ(pthread_mutex_unlock(&fco->mtx));

	fellow_cache_obj_fini(fco);

	AN(fbio);
	fbio->type = FBIO_SEG;
	fbio->u.fcs = fcs;

	fellow_busy_io_submit(fbio);
}

/*
 * ============================================================
 * ASYNC
 *
 */

// handle completed io based on info
static void
fellow_cache_read_complete(struct fellow_cache *fc, void *ptr, int32_t result)
{
	enum fcos_state fcos_next = (typeof(fcos_next))FCOS_CHECK;
	struct fellow_cache_seg *fcs;
	struct fellow_cache_obj *fco;
	unsigned refcount;

	CAST_OBJ_NOTNULL(fcs, ptr, FELLOW_CACHE_SEG_MAGIC);
	assert(FCOS(fcs->state) == FCOS_READING);

	assert(fcs->alloc.size <= INT32_MAX);
	if (result < (int32_t)fcs->alloc.size) {
		if (fc->tune->ioerr_obj == 0)
			FC_WRONG("read error result %d size %d",
			    result, (int32_t)fcs->alloc.size);
		fcos_next = (typeof(fcos_next))FCOS_READFAIL;
		// XXX handle
		INCOMPL();
	}

	//lint -e{655} bit-wise operation uses (compatible) enum's
	fcos_next |= (typeof(fcos_next))FCOS_HIGH(fcs->state);

	fco = fcs->fco;
	CHECK_OBJ_NOTNULL(fco, FELLOW_CACHE_OBJ_MAGIC);

	assert_fcos_transition(fcs->state, fcos_next);

	//lint --e{456} Two execution paths are being combined...
	//lint --e{454} A thread mutex has been locked...
	AZ(pthread_mutex_lock(&fco->mtx));
	fellow_cache_seg_transition_locked(fc, fcs, fcs->state, fcos_next);
	// io holds a ref on the seg and the fco
	if (fellow_cache_seg_deref_locked(fc, fcs))
		AZ(pthread_cond_broadcast(&fco->cond));
	refcount = fellow_cache_obj_deref_locked(fc, fco);
	if (refcount == 0)
		fellow_cache_obj_free(fc, &fco);
	else
		AZ(pthread_mutex_unlock(&fco->mtx));
}

// handle completed io based on info
static void
fellow_cache_async_write_complete(struct fellow_cache *fc,
    void *ptr, int32_t result)
{
	struct fellow_busy_io *fbio;
	struct fellow_busy *fbo;
	struct fellow_cache_seg *fcs = NULL;
	struct fellow_cache_obj *fco;
	enum fcos_state fcos_next = FCOS_INVAL;
	enum fellow_busy_io_e type;
	uint8_t io_outstanding;
	unsigned refcount;

	CHECK_OBJ_NOTNULL(fc, FELLOW_CACHE_MAGIC);
	CHECK_OBJ_NOTNULL(fc->tune, STVFE_TUNE_MAGIC);

	CAST_OBJ_NOTNULL(fbio, ptr, FELLOW_BUSY_IO_MAGIC);

	if (result == -EAGAIN && fbio->retries++ < FBIO_MAX_RETRIES) {
		fellow_busy_io_submit(fbio);
		return;
	}

	type = fbio->type;

	fbo = fbio->fbo;
	CHECK_OBJ_NOTNULL(fbo, FELLOW_BUSY_MAGIC);

	fco = fbo->fco;
	CHECK_OBJ_NOTNULL(fco, FELLOW_CACHE_OBJ_MAGIC);

	if (type == FBIO_SEG) {
		fcs = fbio->u.fcs;
		CHECK_OBJ_NOTNULL(fcs, FELLOW_CACHE_SEG_MAGIC);
		assert(fcs->fco == fco);

		assert(fcs->alloc.size <= INT32_MAX);
#ifdef TODO_WRITE_ERR
		if (FC_INJ || result < (int32_t)fcs->alloc.size) {
			if (fc->tune->ioerr_obj == 0)
				FC_WRONG("write error result %d size %d",
				    result, (int32_t)fcs->alloc.size);
			// XXX handle
			INCOMPL();
		}
#else
		if (result < (int32_t)fcs->alloc.size) {
			FC_WRONG("write error result %d size %d",
			    result, (int32_t)fcs->alloc.size);
		}
#endif

		assert(FCOS(fcs->state) == FCOS_WRITING);
		fcos_next = (typeof(fcos_next))
		    (FCOS_HIGH(fcs->state) | FCOS_INCORE);
		assert_fcos_transition(fcs->state, fcos_next);
	} else {
		assert(type == FBIO_SEGLIST);
		if (result < (int32_t)fbio->u.seglist.reg.size) {
			if (fc->tune->ioerr_obj == 0)
				FC_WRONG("write error result %d size %ld",
				    result, fbio->u.seglist.reg.size);
			// XXX handle
			INCOMPL();
		}
	}

	memset(fbio, 0, sizeof *fbio);

	/* we keep the FCO writing until all else is written
	 * to avoid a race with evict, which removes the oc, which
	 * we need for writing the log
	 *
	 * the writer acquired a ref, so we deref by one
	 *
	 */
	AZ(pthread_mutex_lock(&fco->mtx));
	AN(fbo->io_outstanding);
	io_outstanding = --fbo->io_outstanding;
	if (type == FBIO_SEG) {
		AN(fcs);
		if (FCOS_HIGH(fcs->state) == FCS_HIGH) {
			/* transition a segment */
			assert(fcos_next != FCOS_INVAL);
			fellow_cache_seg_transition_locked(fc, fcs,
			    fcs->state, fcos_next);
			if (fellow_cache_seg_deref_locked(fc, fcs))
				AZ(pthread_cond_broadcast(&fco->cond));
		}
	}
	//lint --e{456} Two execution paths are being combined...
	//lint --e{454} A thread mutex has been locked...
	if (io_outstanding == 0) {
		/* XXX if error, make object mem only and
		 * free disk space
		 */

		switch (fco->logstate) {
		case FCOL_WANTLOG:
			/* XXX could probably this outside the lock
			 * by deleting the obj again if the logstate
			 * was changed to FCOL_TOOLATE
			 */
			fellow_busy_log_submit(fbo);
			stvfe_oc_log_submitted(fco->oc);
			fco->logstate = FCOL_INLOG;
			break;
		case FCOL_NOLOG:
		case FCOL_TOOLATE:
			break;
		case FCOL_DUNNO:
		case FCOL_INLOG:
		case FCOL_DELETED:
			FC_WRONG("fellow_cache_async_write_complete "
			    "wrong logstate %d", fco->logstate);
		default:
			WRONG("fellow_cache_async_write_complete "
			    "invalid logstate");
		}

		fcs = FCO_FCS(fco);
		assert(fcs->state == FCO_WRITING);
		fcos_next = FCO_INCORE;

		/* transition the FCO */
		fellow_cache_seg_transition_locked(fc, fcs,
		    fcs->state, fcos_next);
		refcount = fellow_cache_obj_deref_locked(fc, fco);
		if (refcount == 0) {
			fellow_cache_obj_free(fc, &fco);
		}
		else {
			AZ(pthread_cond_broadcast(&fco->cond));
			AZ(pthread_mutex_unlock(&fco->mtx));
		}
	} else {
		AZ(pthread_mutex_unlock(&fco->mtx));
	}

	if (io_outstanding)
		return;

	fellow_busy_free(&fbo);
}

static void
fellow_cache_seg_async_compl_cb(void *priv,
    /*lint -e{818}*/ struct fellow_io_status *status, unsigned n)
{
	struct fellow_cache *fc;
	unsigned u;

	CAST_OBJ_NOTNULL(fc, priv, FELLOW_CACHE_MAGIC);
	//lint -e{455} not locked mutex unlocked
	AZ(pthread_mutex_unlock(&fc->async_mtx));

	for (u = 0; u < n; u++) {
		switch (faio_info_type(status[u].info)) {
		case FAIOT_CACHE_READ:
			fellow_cache_read_complete(fc,
			    faio_info_ptr(status[u].info),
			    status[u].result);
			break;
		case FAIOT_CACHE_WRITE:
			fellow_cache_async_write_complete(fc,
			    faio_info_ptr(status[u].info),
			    status[u].result);
			break;
		default:
			WRONG("faio_info_type in fellow_cache");
		}
	}
	//lint --e{454} mutex locked but not unlocked
	AZ(pthread_mutex_lock(&fc->async_mtx));
}

static inline struct fellow_disk_seg *
fellow_cache_seg_io_check(const struct fellow_cache *fc,
    struct fellow_cache_seg *fcs)
{
	struct fellow_disk_seg *fds;
	size_t asz;

	CHECK_OBJ_NOTNULL(fcs, FELLOW_CACHE_SEG_MAGIC);
	fds = fcs->disk_seg;
	CHECK_OBJ_NOTNULL(fds, FELLOW_DISK_SEG_MAGIC);

	asz = fellow_rndup(fc->ffd, fds->seg.size);

	assert(fcs->alloc.size == asz);
	AN(fds->seg.off);
	AN(fds->seg.size);
	assert(fds->seg.size >= fcs->len);
	assert(fds->seg.size <= asz);
	fcs->len = fds->seg.size;

	// O_DIRECT alignment on 4K
	AZ((uintptr_t) fcs->alloc.ptr & FELLOW_BLOCK_ALIGN);
	AZ((uintptr_t) fcs->alloc.size & FELLOW_BLOCK_ALIGN);
	AZ((uintptr_t)fds->seg.off & FELLOW_BLOCK_ALIGN);

	return (fds);
}

static void
fellow_cache_seg_sync_read(struct fellow_cache *fc,
    struct fellow_cache_seg * const *segs, unsigned n)
{
	struct fellow_cache_seg *fcs;
	struct fellow_disk_seg *fds;

	while (n > 0) {
		fcs = *segs;
		fds = fellow_cache_seg_io_check(fc, fcs);

		CHECK_OBJ_NOTNULL(fds, FELLOW_DISK_SEG_MAGIC);
		assert(FCOS_IS(fcs->state, READING));
		assert(fds->seg.off >= 0);
		fellow_cache_read_complete(fc, fcs,
		    fellow_io_pread_sync(fc->ffd,
			fcs->alloc.ptr, fcs->alloc.size, fds->seg.off));
		n--;
		segs++;
	}
}

/* no fcs locking required, it needs to be referenced and have the
 * reading/writing state, so can't go away
 */
static void
fellow_cache_seg_async_read(struct fellow_cache *fc,
    struct fellow_cache_seg * const *segs, unsigned n)
{
	struct fellow_cache_seg *fcs;
	struct fellow_disk_seg *fds;
	uint64_t info;
	int r;

	AZ(pthread_mutex_lock(&fc->async_mtx));
	while (n > 0) {
		fcs = *segs;
		info = faio_ptr_info(FAIOT_CACHE_READ, fcs);
		fds = fellow_cache_seg_io_check(fc, fcs);

		CHECK_OBJ_NOTNULL(fds, FELLOW_DISK_SEG_MAGIC);
		assert(FCOS_IS(fcs->state, READING));
		r = fellow_io_read_async_enq(fc->async_ioctx,
		    info, fcs->alloc.ptr, fcs->alloc.size, fds->seg.off);
		if (! r)
			break;
		n--;
		segs++;
	}

	if (fc->async_idle)
		AZ(pthread_cond_signal(&fc->async_cond));
	AZ(pthread_mutex_unlock(&fc->async_mtx));

	if (n == 0)
		return;

	// could not submit, fall back to sync
	DBG("ASYNC FULL; sync fallback %u", n);
	fellow_cache_seg_sync_read(fc, segs, n);
}

static void *
fellow_cache_async_thread(void *priv)
{
	struct fellow_cache *fc;
	unsigned n;

	CAST_OBJ_NOTNULL(fc, priv, FELLOW_CACHE_MAGIC);
	unsigned entries = fellow_io_entries(fc->async_ioctx);

	AZ(pthread_mutex_lock(&fc->async_mtx));
	while (fc->running) {
		struct fellow_io_status status[entries];
		n = fellow_io_wait_completions(fc->async_ioctx,
		    status, entries, 1, fellow_cache_seg_async_compl_cb, fc);
		if (n == 0) {
			fc->async_idle = 1;
			AZ(pthread_cond_wait(&fc->async_cond, &fc->async_mtx));
			fc->async_idle = 0;
		}
	}
	AZ(pthread_mutex_unlock(&fc->async_mtx));
	return (NULL);
}

// XXX better way
#ifndef TEST_DRIVER
static inline void
fellow_cache_seg_evict_locked(struct fellow_cache *fc,
    struct fellow_cache_seg *fcs, struct buddy_ptr_extent *alloc)
{
	AZ(fcs->refcnt);
	AN(fcs->fcs_onlru);
	assert(fcs->state == FCS_INCORE);
	*alloc = fcs->alloc;
	fcs->alloc = buddy_ptr_extent_nil;
	fcs->len = 0;
	fcs->state = FCS_DISK;
	fcs->fcs_onlru = 0;
	VTAILQ_REMOVE(&fc->lru_head, fcs, lru_list);
}

static int
fellow_cache_lru_work(struct worker *wrk, struct fellow_cache *fc)
{
	buddy_t *buddy = fc->membuddy;
	struct fellow_cache_seg *fcs, *fcss;
	struct fellow_cache_obj *fco;
	struct objcore *oc;
	struct buddy_ptr_extent alloc;
	int r;

	//lint --e{456} flexelint does not grok trylock

	alloc = buddy_ptr_extent_nil;
	oc = NULL;
	fco = NULL;
	AZ(pthread_mutex_lock(&fc->lru_mtx));
	VTAILQ_FOREACH_SAFE(fcs, &fc->lru_head, lru_list, fcss) {
		// no use trying the same object again and again
		if (fcs->fco == fco)
			continue;
		fco = fcs->fco;
		AN(fco);
		r = pthread_mutex_trylock(&fco->mtx);
		if (r != 0) {
			assert(r == EBUSY);
			continue;
		}
		if (fcs->state == FCS_INCORE) {
			fellow_cache_seg_evict_locked(fc, fcs,
			    &alloc);
			AZ(pthread_mutex_unlock(&fco->mtx));
			fco = NULL;
			break;
		}
		if (fcs->state == FCO_INCORE) {
			oc = fco->oc;
			CHECK_OBJ_NOTNULL(oc, OBJCORE_MAGIC);
			r = stvfe_mutate(wrk, oc);
			AZ(pthread_mutex_unlock(&fco->mtx));
			if (r) {
				/* mutate was successful
				 * VSC_C_main->n_lru_nuked++; // XXX per lru ?
				 */
				break;
			}
			oc = NULL;
			fco = NULL;
			continue;
		}
		AZ(pthread_mutex_unlock(&fco->mtx));
		WRONG("fcs state in lru");
	}
	AZ(pthread_mutex_unlock(&fc->lru_mtx));

	if (fcs == NULL) {
//		VSLb(wrk->vsl, SLT_ExpKill, "LRU_Fail");
		AZ(alloc.ptr);
		AZ(oc);
		return (0);
	}

	if (alloc.ptr != NULL) {
		AZ(fco);
		AN(alloc.size);
		buddy_return1_ptr_extent(buddy, &alloc);
	} else if (fco != NULL) {
		AN(oc);
//		VSLb(wrk->vsl, SLT_ExpKill, "LRU x=%u", ObjGetXID(wrk, oc));

		// we removed the oc's reference on fco in stvfe_mutate()
		fellow_cache_obj_deref(fc, fco);
	} else {
		INCOMPL();
	}
	return (1);
}

static void
reserve_free(buddy_t *buddy, struct buddy_ptr_page *r, unsigned n)
{
	struct buddy_returns *rets =
	    BUDDY_RETURNS_STK(buddy, BUDDY_RETURNS_MAX);

	if (n == 0)
		return;

	AN(r);

	while (n--)
		AN(buddy_return_ptr_page(rets, r++));

	buddy_return(rets);
}

static void
reserve_req(struct buddy_reqs *reqs, unsigned n, unsigned bits)
{
	if (n > BUDDY_REQS_MAX)
		n = BUDDY_REQS_MAX;
	while (n--)
		AN(buddy_req_page(reqs, bits, 0));
}

static void
reserve_fill(struct buddy_ptr_page *r, const struct buddy_reqs *reqs, uint8_t n)
{
	uint8_t u;

	for (u = 0; u < n; u++) {
		AZ(r->ptr);
		*r++ = buddy_get_ptr_page(reqs, u);
	}
}

void *
fellow_cache_lru_thread(struct worker *wrk, void *priv)
{
	struct vsl_log vsl;
	struct buddy_reqs *reqs;
	struct fellow_cache *fc;
	buddy_t *buddy;
	struct buddy_ptr_page *r = NULL;
	unsigned i, filled = 0, nr = 0, rc, cb;
	uint8_t n;
	size_t sz;

	CAST_OBJ_NOTNULL(fc, priv, FELLOW_CACHE_MAGIC);
	buddy = fc->membuddy;
	CHECK_OBJ(buddy, BUDDY_MAGIC);

	// fellow_cache_seg_evict_locked()
	assert_fcos_transition(FCS_INCORE, FCS_DISK);

	reqs = BUDDY_REQS_STK(buddy, BUDDY_REQS_MAX);
	BUDDY_REQS_PRI(reqs, FEP_RESERVE);

	AZ(wrk->vsl);
	wrk->vsl = &vsl;
	VSL_Setup(wrk->vsl, NULL, (size_t)0);

	while (fc->running) {
		cb = fc->tune->chunk_exponent;
		rc = fc->tune->mem_reserve_chunks;

		// reserve pointers
		while (nr != rc) {
			if (nr != 0)
				AN(r);
			reserve_free(buddy, r, filled);
			filled = 0;
			nr = rc;
#ifdef LRU_NOISE
			VSLb(wrk->vsl, SLT_Debug,
			    "sfedsk reserve config cb=%u nr=%u",
			    cb, nr);
#endif
			if (nr == 0)
				break;

			sz = nr * sizeof *r;
			r = realloc(r, sz);
			AN(r);
			memset(r, 0, sz);
			break;
		};

		// fill reserve
		n = 1;
		while (filled < nr && buddy->waiting == 0 && n > 0) {
			AN(r);
			reserve_req(reqs, nr - filled, cb);
			(void) BUDDYF(alloc_async)(reqs);

			while (buddy->waiting)
				if (! fellow_cache_lru_work(wrk, fc))
					break;

			n = BUDDYF(alloc_async_ready)(reqs);
			assert(filled + n <= nr);
			reserve_fill(r + filled, reqs, n);
			filled += n;
			BUDDYF(alloc_async_done)(reqs);
		}

#ifdef LRU_NOISE
		if (nr)
			VSLb(wrk->vsl, SLT_Debug,
			    "sfemem reserve fill: %u", filled);
#endif

		VSL_Flush(wrk->vsl, 0);
		stvfe_sumstat(wrk);
		buddy_wait_needspace(buddy);

		// drain reserve
		while (filled > 0 && buddy->waiting > 0) {
			i = --filled;

			AN(r);
			AN(r[i].ptr);
			BUDDYF(return1_ptr_page)(buddy, &r[i]);
		}

#ifdef LRU_NOISE
		if (nr)
			VSLb(wrk->vsl, SLT_Debug,
			    "sfemem reserve use : %u left", filled);
#endif

		while (buddy->waiting)
			if (! fellow_cache_lru_work(wrk, fc)) {
				/* give other threads a change to
				 * get the lru lock */
				(void)usleep(10*1000);
			}
	}
	reserve_free(buddy, r, filled);
	free(r);
	return (NULL);
}
#endif

/* returns if moved */
int
fellow_cache_obj_lru_touch(struct fellow_cache *fc,
    struct fellow_cache_obj *fco)
{
	struct fellow_cache_seg *fcs;
	int r;

	CHECK_OBJ_NOTNULL(fc, FELLOW_CACHE_MAGIC);
	CHECK_OBJ_NOTNULL(fco, FELLOW_CACHE_OBJ_MAGIC);
	fcs = FCO_FCS(fco);
	CHECK_OBJ_NOTNULL(fcs, FELLOW_CACHE_SEG_MAGIC);

	if (! fcs->fcs_onlru)
		return (0);

	assert(fellow_cache_shouldlru(fcs->state, fco->oc, fcs->refcnt));

	r = pthread_mutex_trylock(&fc->lru_mtx);
	if (r != 0) {
		assert(r == EBUSY);
		return (0);
	}
	VTAILQ_REMOVE(&fc->lru_head, fcs, lru_list);
	VTAILQ_INSERT_TAIL(&fc->lru_head, fcs, lru_list);
	//lint -e{455} flexelint does not grok trylock
	AZ(pthread_mutex_unlock(&fc->lru_mtx));

	return (1);
}

// not unter logmtx
static void
fellow_cache_async_init(struct fellow_cache *fc, fellow_task_run_t taskrun)
{
	unsigned entries;

	CHECK_OBJ_NOTNULL(fc, FELLOW_CACHE_MAGIC);

	entries = fellow_io_ring_size("fellow_cache_io_entries");
	DBG("io entries %u", entries);

	AZ(pthread_mutex_init(&fc->async_mtx, NULL));
	AZ(pthread_cond_init(&fc->async_cond, NULL));
	fc->async_ioctx = fellow_io_init(fellow_fd(fc->ffd),
	    entries, fc->membuddy->area, fc->membuddy->map->size,
	    taskrun);
	AN(fc->async_ioctx);
	AZ(pthread_create(&fc->async_thread, NULL,
		fellow_cache_async_thread, fc));
}

static void
fellow_cache_async_fini(struct fellow_cache *fc)
{
	CHECK_OBJ_NOTNULL(fc, FELLOW_CACHE_MAGIC);

	AZ(pthread_mutex_lock(&fc->async_mtx));
	AZ(pthread_cond_signal(&fc->async_cond));
	AZ(pthread_mutex_unlock(&fc->async_mtx));
	AZ(pthread_join(fc->async_thread, NULL));
	AZ(pthread_cond_destroy(&fc->async_cond));
	AZ(pthread_mutex_destroy(&fc->async_mtx));
}


/* .
 * .
 * .
 * ASYNC END
 * ============================================================
 */

static void
fellow_cache_seg_deref(struct fellow_cache *fc,
    struct fellow_cache_seg * const *segs, unsigned n)
{
	struct fellow_cache_seg *fcs;
	struct fellow_cache_obj *fco;

	CHECK_OBJ_NOTNULL(*segs, FELLOW_CACHE_SEG_MAGIC);
	AN(n);
	fco = (*segs)->fco;
	CHECK_OBJ_NOTNULL(fco, FELLOW_CACHE_OBJ_MAGIC);

	AZ(pthread_mutex_lock(&fco->mtx));
	while (n--) {
		fcs = *segs++;
		CHECK_OBJ_NOTNULL(fcs, FELLOW_CACHE_SEG_MAGIC);
		(void) fellow_cache_seg_deref_locked(fc, fcs);
	}
	AZ(pthread_mutex_unlock(&fco->mtx));
}

static const char *
fellow_cache_seg_check(struct fellow_cache *fc, struct fellow_cache_seg *fcs)
{
	const struct fellow_disk_seg *fds;
	struct fellow_cache_obj *fco;
	enum fcos_state to = FCS_INCORE;
	const char *err = NULL;

	switch (fcs->state) {
	case FCS_READFAIL:
		return ("segment FCS_READFAIL");
	case FCS_WRITING:
	case FCS_INCORE:
		return (NULL);
	case FCS_CHECK:
		break;
	default:
		WRONG("segment state in _check");
	}

	fds = fcs->disk_seg;
	CHECK_OBJ_NOTNULL(fds, FELLOW_DISK_SEG_MAGIC);
#ifdef TODO_READ_ERR
	if ((fds->segnum % 16 == 0 && FC_INJ) ||
	    fhcmp(fds->fht, fds->fh, fcs->alloc.ptr, fcs->len)) {
		err = "segment checksum error";
		to = FCS_READFAIL;
	}
#endif

	fco = fcs->fco;
	AZ(pthread_mutex_lock(&fco->mtx));
	if (fcs->state == FCS_CHECK) {
		// re-check, could have raced
		fellow_cache_seg_transition_locked(fc, fcs,
		    FCS_CHECK, to);
		AZ(pthread_cond_broadcast(&fco->cond));
	}
	AZ(pthread_mutex_unlock(&fco->mtx));

	return (err);
}

enum fellow_cache_io_e {
	FCIO_SYNC,
	FCIO_ASYNC
};

/* for potentially multiple segments from the same fco, grab references and read
 * them
 */

static void
fellow_cache_seg_ref_in(struct fellow_cache *fc, enum fellow_cache_io_e type,
    struct fellow_cache_seg * const *segs, const unsigned n)
{
	struct fellow_cache_seg *fcs, *iosegs[n], *racesegs[n];
	struct fellow_cache_obj *fco;
	unsigned u, ion = 0, racen = 0;
	unsigned req2seg[n];
	uint8_t r;
	struct buddy_ptr_extent mem[n];
	struct buddy_reqs *reqs;
	struct buddy_returns *rets;

	CHECK_OBJ_NOTNULL(*segs, FELLOW_CACHE_SEG_MAGIC);
	AN(n);
	fco = (*segs)->fco;
	CHECK_OBJ_NOTNULL(fco, FELLOW_CACHE_OBJ_MAGIC);

	assert(n <= BUDDY_REQS_MAX);
	reqs = BUDDY_REQS_STK(fc->membuddy, BUDDY_REQS_MAX);
	BUDDY_REQS_PRI(reqs, FEP_NORMAL);
	rets = BUDDY_RETURNS_STK(fc->membuddy, BUDDY_RETURNS_MAX);

	memset(iosegs, 0, sizeof iosegs);
	memset(racesegs, 0, sizeof racesegs);
	memset(req2seg, 0, sizeof req2seg);
	memset(mem, 0, sizeof mem);

	/* opportunistically alloc memory and conduct checks on the way */
	for (r = 0, u = 0; u < n; u++) {
		struct fellow_disk_seg *fds;

		fcs = segs[u];

		CHECK_OBJ_NOTNULL(fcs, FELLOW_CACHE_SEG_MAGIC);
		fds = fcs->disk_seg;
		CHECK_OBJ_NOTNULL(fds, FELLOW_DISK_SEG_MAGIC);
		assert(fcs->fco == fco);

		switch (fcs->state) {
		case FCS_DISK:
			req2seg[r++] = u;
			AN(buddy_req_extent(reqs,
				fellow_rndup(fc->ffd, fds->seg.size), 0));
			break;
		default:
			break;
		}

		if (r != BUDDY_REQS_MAX)
			continue;

		r = buddy_alloc_wait(reqs);
		assert(r == BUDDY_REQS_MAX);
		while (r--) {
			AZ(mem[req2seg[r]].ptr);
			mem[req2seg[r]] = buddy_get_ptr_extent(reqs, r);
		}
		buddy_alloc_wait_done(reqs);
		r = 0;
	}

	r = buddy_alloc_wait(reqs);
	while (r--) {
		AZ(mem[req2seg[r]].ptr);
		mem[req2seg[r]] = buddy_get_ptr_extent(reqs, r);
	}
	buddy_alloc_wait_done(reqs);
	r = 0;

	/* grab references, remember segments which need io and those for which
	 * we raced memory alloc
	 */

	AZ(pthread_mutex_lock(&fco->mtx));

	for (u = 0; u < n; u++) {
		fcs = segs[u];

		(void) fellow_cache_seg_ref_locked(fc, fcs);

		while (type == FCIO_SYNC &&
		    (fcs->state == FCS_BUSY || fcs->state == FCS_READING))
			fellow_cache_seg_wait_locked(fcs);

		switch (fcs->state) {
		case FCS_BUSY:
		case FCS_READING:
			assert(type == FCIO_ASYNC);
			break;
		case FCS_READFAIL:
			INCOMPL();
			break;
		case FCS_WRITING:
		case FCS_CHECK:
		case FCS_INCORE:
			break;
		case FCS_DISK:
			if (mem[u].ptr == NULL) {
				/*
				 * race exception (should be rare).  we do not
				 * want to malloc under the fco mtx, so we
				 * restart later
				 */
				(void) fellow_cache_seg_deref_locked(fc, fcs);
				racesegs[racen++] = fcs;
				break;
			}

			// reference for io
			(void) fellow_cache_seg_ref_locked(fc, fcs);

			fcs->alloc = mem[u];
			mem[u] = buddy_ptr_extent_nil;

			fellow_cache_seg_transition_locked_notincore(fcs,
			    FCS_READING);

			iosegs[ion++] = fcs;
			break;
		default:
			WRONG("cache_seg_in state");
		}
	}
	/*
	 * when we get here, the fco must already be ref'd
	 * each io get a reference on the fco also, plus
	 */
	AN(FCO_REFCNT(fco));
	FCO_REFCNT(fco) += ion;
	AZ(pthread_mutex_unlock(&fco->mtx));

	/* free unused allocations */
	for (u = 0; u < n; u++) {
		if (mem[u].ptr == NULL)
			continue;
		AN(buddy_return_ptr_extent(rets, &mem[u]));
	}
	buddy_return(rets);

	/* issue ios */
	switch (type) {
	case FCIO_SYNC:
		fellow_cache_seg_sync_read(fc, iosegs, ion);
		break;
	case FCIO_ASYNC:
		fellow_cache_seg_async_read(fc, iosegs, ion);
		break;
	default:
		WRONG("io type");
	}

	/* retry any raced */
	if (racen > 0)
		fellow_cache_seg_ref_in(fc, type, racesegs, racen);
}

static inline int
fcs_state_read_needwait(enum fcos_state s)
{
	switch (s) {
	case FCS_BUSY:
	case FCS_READING:
		return (1);
	case FCS_CHECK:
	case FCS_WRITING:
	case FCS_INCORE:
	case FCS_READFAIL:
	case FCS_USABLE:
		return (0);
	default:
		// INIT, DISK
		WRONG("state when waiting for read");
	}
}

struct fellow_cache_res
fellow_cache_obj_iter(struct fellow_cache *fc, struct fellow_cache_obj *fco,
    void *priv, objiterate_f func, int final)
{
	CHECK_OBJ_NOTNULL(fc, FELLOW_CACHE_MAGIC);
	CHECK_OBJ_NOTNULL(fc->tune, STVFE_TUNE_MAGIC);

	struct fellow_cache_res fcr;
	unsigned readahead = fc->tune->readahead;
	unsigned mod = readahead + 1;
	struct fellow_cache_seg *fcsnext, *fcs, *fcsra,
	    *ra[mod], *raio[mod], *deref[mod];
	struct fcscursor c, rac, nextc;
	unsigned n = 0, ran = 0, raion, derefn = 0, flags, flush;
	const char *err;
	int ret2;

	fcr.status = fcr_ok;	// also if func() != 0
	fcr.r.integer = 0;

	CHECK_OBJ_NOTNULL(fco, FELLOW_CACHE_OBJ_MAGIC);

	memset(ra, 0, sizeof ra);
	memset(raio, 0, sizeof raio);
	memset(deref, 0, sizeof deref);

	fcsc_init(&c, &fco->seglist);
	rac = c;
	(void) fcsc_next(&rac);
	nextc = rac;

	flags = final ? OBJ_ITER_FLUSH : 0;
	flush = 0;
	while ((fcs = fcsc_next(&c)) != NULL) {
		if (ra[n % mod] == NULL) {
			/* We always flush for readahead == 0, see below for
			 * explanation */
			AZ(derefn);
			fellow_cache_seg_ref_in(fc, FCIO_SYNC, &fcs, 1);
			ra[n % mod] = fcs;
		}
		assert(ra[n % mod] == fcs);

		raion = 0;
		while (ran < n + readahead) {
			fcsra = fcsc_next(&rac);
			if (fcsra == NULL)
				break;
			ran++;
			raio[raion++] = fcsra;
			TAKEZN(ra[ran % mod], fcsra);
		}
		if (raion)
			fellow_cache_seg_ref_in(fc, FCIO_ASYNC,
			    raio, raion);

		AZ(fcr.r.integer);
		if (fcs_state_read_needwait(fcs->state)) {
			/* We can not get here for readahead == 0, because of
			 * the _ref_in(FCIO_SYNC) above
			 *
			 * last round, we checked if this fcs was in core and
			 * flushed if it was not. Because we hold a reference,
			 * it can not go back from INCORE/CHECK to BUSY or
			 * READING, so the deref list needs to be empty
			 */
			AN(readahead);
			AZ(derefn);

			AZ(pthread_mutex_lock(&fcs->fco->mtx));
			while (fcs_state_read_needwait(fcs->state))
				fellow_cache_seg_wait_locked(fcs);
			AZ(pthread_mutex_unlock(&fcs->fco->mtx));
		}

		// aborted busy segment
		if (fcs->state == FCS_USABLE)
			break;

		err = fellow_cache_seg_check(fc, fcs);
		if (err != NULL) {
			fcr = FCR_IOFAIL(err);
			break;
		}

		assert(fcs->state != FCS_READFAIL);

		/* peek at next segment for END
		 * flush if next segment not in ram or no deref space left
		 *
		 * for readahead == 0, this implies flushing always because
		 * derefn == mod - 1 == 0
		 */
		AZ(flags & OBJ_ITER_END);
		fcsnext = fcsc_next(&nextc);
		flush = 0;
		if (fcsnext == NULL || fcsnext->state == FCS_USABLE) {
			flags |= OBJ_ITER_END;
			flush = OBJ_ITER_FLUSH;
		}
		else if (derefn == mod - 1 || fcsnext->state != FCS_INCORE)
			flush = OBJ_ITER_FLUSH;

		assert(ra[n % mod] == fcs);
		ra[n % mod] = NULL;

		fcr.r.integer = func(priv, flags | flush,
		    fcs->alloc.ptr, (ssize_t)fcs->disk_seg->seg.size);

		if (final) {
			AZ(derefn);
			AN(flags & OBJ_ITER_FLUSH);
			/* if the opportunistic free fails, the segment will
			 * get deleted when the object is */
			AZ(pthread_mutex_lock(&fcs->fco->mtx));
			if (fcs->refcnt == 1)
				fellow_cache_seg_free(fc, fcs, 1);
			else
				AN(fellow_cache_seg_deref_locked(fc, fcs));
			AZ(pthread_mutex_unlock(&fcs->fco->mtx));
		} else if (flush) {
			assert(derefn < mod);
			deref[derefn++] = fcs;
			fellow_cache_seg_deref(fc, deref, derefn);
			derefn = 0;
		} else {
			assert(derefn < mod);
			deref[derefn++] = fcs;
		}
		assert(fcr.status == fcr_ok);
		if (fcr.r.integer)
			break;
		n++;
	}

	if (fcr.status != fcr_ok || fcr.r.integer != 0)
		while (mod--) {
			if (ra[mod] != NULL)
				fellow_cache_seg_deref(fc, &ra[mod], 1);
		}
	else
		while (mod--)
			AZ(ra[mod]);

	if (derefn)
		fellow_cache_seg_deref(fc, deref, derefn);

	if ((flags & OBJ_ITER_END) == 0) {
		ret2 = func(priv, OBJ_ITER_END, NULL, (size_t)0);
		if (fcr.status == fcr_ok && fcr.r.integer == 0)
			fcr.r.integer = ret2;
	}

	return (fcr);
}

static struct fellow_cache_res
fellow_busy_body_seg_next(struct fellow_busy *fbo)
{
	struct fellow_cache_seglist *fcsl;
	struct fellow_cache_seg *fcs;
	struct fellow_cache_res fcr;

	CHECK_OBJ_NOTNULL(fbo, FELLOW_BUSY_MAGIC);
	fcsl = fbo->body_seglist;

	AN(fcsl->fdsl);
	assert(fcsl->fdsl->nsegs <= fcsl->lsegs);
	if (fcsl->fdsl->nsegs == fcsl->lsegs ||
	    fcsl->segs[fcsl->fdsl->nsegs].state == FCS_INIT) {
		fcr = fellow_body_seglist_alloc(fbo, fbo->body_seglist);
		if (fcr.status != fcr_ok)
			return (fcr);
		CAST_OBJ_NOTNULL(fcsl, fcr.r.ptr, FELLOW_CACHE_SEGLIST_MAGIC);
		fbo->body_seglist = fcsl;
	}

	AN(fcsl->fdsl);
	assert(fcsl->fdsl->nsegs < fcsl->lsegs);

	fcs = &fcsl->segs[fcsl->fdsl->nsegs];
	fbo->body_seg = fcs;
	assert(fcs->state == FCS_USABLE);
	assert(fcs->fco == fbo->fco);
	AZ(fcs->refcnt);
	return (FCR_OK(fcs));
}

/*
 * segment already has disk space allocated, allocate memory
 * half way between FCS_USABLE and FCS_BUSY, returns BUSY
 *
 * used for body and auxattr
 */
static size_t fellow_busy_seg_memalloc(buddy_t *membuddy,
    struct fellow_cache_seg *fcs)
{
	struct fellow_disk_seg *fds;

	CHECK_OBJ_NOTNULL(fcs, FELLOW_CACHE_SEG_MAGIC);
	fds = fcs->disk_seg;
	CHECK_OBJ_NOTNULL(fds, FELLOW_DISK_SEG_MAGIC);

	assert (fcs->state == FCS_USABLE);
	AN(fds->seg.off);
	AN(fds->seg.size);

	AZ(fcs->alloc.ptr);
	AZ(fcs->alloc.size);

	fcs->alloc = buddy_alloc1_ptr_extent_wait(membuddy, FEP_NORMAL,
	    fds->seg.size, 0);

	if (fcs->alloc.ptr == NULL)
		return (0);

	memset(fcs->alloc.ptr, 0, fcs->alloc.size);

	// membuddy min < dskbuddy min
	assert(fds->seg.size == fcs->alloc.size);
	AZ(fcs->refcnt);
	fcs->refcnt = 1;
	fellow_cache_seg_transition_locked_notincore(fcs, FCS_BUSY);

	return (fcs->alloc.size);
}

/*
 * segment for auxattr
 */
static int
fellow_busy_seg_alloc(struct fellow_busy *fbo,
    struct fellow_cache_seg *fcs, size_t size)
{
	struct fellow_cache *fc;
	struct buddy_off_extent *fdr;
	struct fellow_disk_seg *fds;

	CHECK_OBJ_NOTNULL(fbo, FELLOW_BUSY_MAGIC);
	fc = fbo->fc;
	CHECK_OBJ_NOTNULL(fc, FELLOW_CACHE_MAGIC);

	assert (fcs->state == FCS_USABLE);
	assert_cache_seg_consistency(fcs);

	fds = fcs->disk_seg;
	CHECK_OBJ_NOTNULL(fds, FELLOW_DISK_SEG_MAGIC);

	fdr = fellow_busy_region_alloc(fbo, size, 0);
	if (fdr == NULL)
		return (0);
	fds->seg = *fdr;
	if (fellow_busy_seg_memalloc(fbo->fc->membuddy, fcs) == 0)
		// we do not need to free because we use the busy region
		return (0);
	return (1);
}


// varnish-cache will ask for content-length if it is known.
// otherwise it will ask for cache_param->fetch_chunksize
// so we only need to limit that for mem
// for disk, we need to multiply by two each time we extend

struct fellow_cache_res
fellow_busy_obj_getspace(struct fellow_busy *fbo, size_t *sz, uint8_t **ptr)
{
	struct fellow_cache_seg *fcs, *unbusy = NULL;
	size_t max, spc, delta;
	struct fellow_cache_res fcr;

	CHECK_OBJ_NOTNULL(fbo, FELLOW_BUSY_MAGIC);
	AN(sz);
	assert(*sz > 0);
	AN(ptr);

	CHECK_OBJ_NOTNULL(fbo->fc, FELLOW_CACHE_MAGIC);
	CHECK_OBJ_NOTNULL(fbo->fc->tune, STVFE_TUNE_MAGIC);
	max = (size_t)1 << fbo->fc->tune->chunk_exponent;

	/*
	 * for content-length, varnish-cache will ask for less and less.
	 *
	 * if it keeps on asking for more, we
	 * increment each time we exceed the previous estimate
	 */

	if (fbo->sz_estimate == 0) {
		fbo->sz_estimate = *sz;
		fbo->sz_increment = *sz;
	} else if (*sz + fbo->sz_returned > fbo->sz_estimate) {
		delta = *sz + fbo->sz_returned - fbo->sz_estimate;
		if (fbo->sz_increment < delta)
			fbo->sz_increment = delta;

		fbo->sz_estimate += fbo->sz_increment;
		fbo->sz_increment <<= 1;
	}

	fcs = fbo->body_seg;
	if (fcs != NULL && fcs->state == FCS_BUSY) {
		AN(fcs->alloc.size);
		if (fcs->len < fcs->alloc.size) {
			spc = fcs->alloc.size - fcs->len;
			assert(spc > 0);
			if (spc < *sz)
				*sz = spc;
			*ptr = (uint8_t *)fcs->alloc.ptr + fcs->len;
			return (FCR_OK(fbo));
		}

		assert(fcs->len == fcs->alloc.size);
		unbusy = fcs;
		fcs = NULL;
	}
	if (fcs == NULL) {
		fcr = fellow_busy_body_seg_next(fbo);
		if (fcr.status != fcr_ok)
			goto out;
		CAST_OBJ_NOTNULL(fcs, fcr.r.ptr, FELLOW_CACHE_SEG_MAGIC);
	}

	assert(fcs->state == FCS_USABLE);
	assert_cache_seg_consistency(fcs);

	spc = *sz;
	if (spc > max)
		spc = max;

	/* we set the state of the next segment to BUSY
	 * before we unbusy the previous to signal to
	 * _iter() that more data is to come
	 */

	spc = fellow_busy_body_seg_alloc(fbo, fcs->disk_seg, spc);
	if (spc == 0) {
		fcr = FCR_ALLOCFAIL("body disk seg alloc");
		goto out;
	}

	spc = fellow_busy_seg_memalloc(fbo->fc->membuddy, fcs);
	if (spc == 0) {
		fellow_busy_body_seg_return(fbo, fcs->disk_seg);
		fcr = FCR_ALLOCFAIL("body memory seg alloc");
		goto out;
	}
	assert(fcs->state == FCS_BUSY);
	fbo->body_seglist->fdsl->nsegs++;
	assert(fbo->body_seglist->fdsl->nsegs <= fbo->body_seglist->lsegs);

	*sz = spc;
	*ptr = fcs->alloc.ptr;

	fcr = FCR_OK(fbo);
  out:
	if (unbusy != NULL)
		fellow_cache_seg_unbusy(fbo, unbusy);

	return (fcr);
}

// caller commits used space
void
fellow_busy_obj_extend(struct fellow_busy *fbo, size_t l)
{
	struct fellow_cache_seg *fcs;

	CHECK_OBJ_NOTNULL(fbo, FELLOW_BUSY_MAGIC);
	assert(l > 0);

	fcs = fbo->body_seg;
	CHECK_OBJ_NOTNULL(fcs, FELLOW_CACHE_SEG_MAGIC);
	assert(fcs->state == FCS_BUSY);
	fcs->len += l;
	fbo->sz_returned += l;
	assert(fcs->len <= fcs->alloc.size);
}

static void
fellow_busy_obj_trim_seglists(struct fellow_busy *fbo)
{
	struct fellow_cache_seglist *fcsl;
	struct fellow_disk_seglist *fdsl;
	struct fellow_cache_obj *fco;

	CHECK_OBJ_NOTNULL(fbo, FELLOW_BUSY_MAGIC);
	fco = fbo->fco;
	CHECK_OBJ_NOTNULL(fco, FELLOW_CACHE_OBJ_MAGIC);

	AN(fbo->body_seglist);
	AN(fbo->body_seglist->fdsl);
	// only to be called if anything to trim at all
	AZ(fbo->body_seglist->fdsl->nsegs);

	fcsl = &fco->seglist;
	CHECK_OBJ(fcsl, FELLOW_CACHE_SEGLIST_MAGIC);

	// not for the embedded seglist
	if (fbo->body_seglist == fcsl)
		return;

	// find the previous seglist
	while (fcsl->next != fbo->body_seglist) {
		fcsl = fcsl->next;
		CHECK_OBJ_NOTNULL(fcsl, FELLOW_CACHE_SEGLIST_MAGIC);
	}
	assert(fcsl->next == fbo->body_seglist);
	fcsl->next = NULL;
	fdsl = fcsl->fdsl;
	fellow_busy_region_free(fbo, &fdsl->next);
	AZ(fdsl->next.off);
	AZ(fdsl->next.size);

	/* not holding fco->mtx because we trim surplus seglists
	 * which are not on the lru
	 */
	fellow_cache_seglist_free(fbo->fc, fbo->body_seglist);
}

/*
 * we need to write full disk blocks and if the membuddy minsize is less than
 * the dskbuddy minsize, we could trim to under the blocksize
 *
 * so we make this a two-step process: Determine surplus, trim disk allocation,
 * trim mem allocation if there is an FCS_BUSY
 */

void
fellow_busy_obj_trimstore(struct fellow_busy *fbo)
{
	struct fellow_body_region *fbr;
	struct buddy_off_extent *fdr;
	struct fellow_cache_seglist *fcsl;
	struct fellow_cache_seg *fcs;
	struct fellow_cache *fc;
	size_t nsz, base, surplus = 0;

	CHECK_OBJ_NOTNULL(fbo, FELLOW_BUSY_MAGIC);
	fcsl = fbo->body_seglist;
	CHECK_OBJ_NOTNULL(fcsl, FELLOW_CACHE_SEGLIST_MAGIC);
	fcs = fbo->body_seg;
	CHECK_OBJ_NOTNULL(fcs, FELLOW_CACHE_SEG_MAGIC);
	fc = fbo->fc;
	CHECK_OBJ_NOTNULL(fc, FELLOW_CACHE_MAGIC);

	fbr = &fbo->body_region;
	fdr = fbr->reg;

	assert(fcs->state == FCS_USABLE || fcs->state == FCS_BUSY);
	assert_cache_seg_consistency(fcs);

	if (fcs->state == FCS_BUSY) {
		assert(fcs->len <= fcs->alloc.size);
		AN(fdr);
	} else {
		assert(fcs->state == FCS_USABLE);
		AZ(fcs->alloc.size);
		AZ(fcs->len);
	}

	if (fdr) {
		assert(fbr->len >= fcs->alloc.size);

		// fbr offset where fcs starts
		base = fbr->len - fcs->alloc.size;
		nsz = base + fcs->len;
		fellow_region_trim(fc->ffd, fdr, nsz);

		nsz = fdr->size;
		if (nsz == 0)
			fbo->nregion--;

		assert(fcs->alloc.size >= (nsz - base));
		surplus = fcs->alloc.size - (nsz - base);
	}

//	DBG("surplus %zu", surplus);

	if (fcs->state == FCS_BUSY) {
		assert(fcs->alloc.size >= surplus);
		nsz = fcs->alloc.size - surplus;

		buddy_trim1_ptr_extent(fc->membuddy, &fcs->alloc, nsz);

		// dsk/membuddy aligment must match
		assert(nsz == fcs->alloc.size);
		fcs->disk_seg->seg.size = fcs->alloc.size;
		assert(fcs->len <= fcs->alloc.size);
		if (fcs->len == 0) {
			fcs->alloc.ptr = NULL;
			AZ(fcs->disk_seg->seg.size);
			fcs->disk_seg->seg.off = 0;
			fellow_cache_seg_transition(fbo->fc, fcs,
			    FCS_BUSY, FCS_USABLE);
			fcsl->fdsl->nsegs--;
			assert(fcs == &fcsl->segs[fcsl->fdsl->nsegs]);
		} else {
			fellow_cache_seg_unbusy(fbo, fcs);
		}
	} else {
		AZ(surplus);
		assert(fcs->state == FCS_USABLE);
	}

	if (fcsl->fdsl->nsegs == 0)
		fellow_busy_obj_trim_seglists(fbo);

	// invalidate
	fbo->body_seg = NULL;
	fbo->body_seglist = NULL;
}

#ifdef XXXDEBUG
static int
fdr_compar(const void *aa, const void *bb)
{
	const struct buddy_off_extent *a = aa;
	const struct buddy_off_extent *b = bb;

	if (a->off < b->off)
		return (-1);
	if (a->off > b->off)
		return (1);
	if (a->size < b->size)
		return (-1);
	if (a->size > b->size)
		return (1);
	return (0);
}
#endif

/* under fco mtx */
static void
fellow_busy_log_submit(const struct fellow_busy *fbo)
{
	unsigned r = 0, ndle = (fbo->nregion / DLE_REG_NREGION) + 2;
	struct fellow_dle *e, dle_obj[1], dles[ndle];
	struct fellow_cache_obj *fco;
	const struct fellow_disk_obj *fdo;

	CHECK_OBJ_NOTNULL(fbo, FELLOW_BUSY_MAGIC);
	fco = fbo->fco;
	CHECK_OBJ_NOTNULL(fco, FELLOW_CACHE_OBJ_MAGIC);
	fdo = FCO_FDO(fco);
	CHECK_OBJ_NOTNULL(fdo, FELLOW_DISK_OBJ_MAGIC);

	AN(fdo->fdo_flags & FDO_F_INLOG);

	fellow_dle_init(dle_obj, 1);
	stvfe_oc_dle_obj(fco->oc, dle_obj);

	fellow_dle_init(dles, ndle);
	if (fbo->nregion)
		r = fellow_dle_reg_fill(dles, ndle, fbo->region, fbo->nregion,
		    DLE_REG_ADD, dle_obj->u.obj.hash);

	e = &dles[r++];
	e->type = DLEDSK(DLE_OBJ_ADD);
	memcpy(&e->u.obj, &dle_obj->u.obj, sizeof e->u.obj);
	e->u.obj.start = fbo->fco->fdb;

	fellow_log_dle_submit(fbo->fc->ffd, dles, r);

#ifdef XXXDEBUG
	struct buddy_off_extent region[FCO_MAX_REGIONS];
	unsigned n;

	n = fellow_obj_regions(fbo->fc->ffd, fbo->fco, region);
	assert(n == fbo->nregion);
	qsort(fbo->region, n, sizeof(struct buddy_off_extent), fdr_compar);
	qsort(region, n, sizeof(struct buddy_off_extent), fdr_compar);
	for (r = 0; r < n; r++) {
		assert(region[r].off == fbo->region[r].off);
		assert(region[r].size == fbo->region[r].size);
	}
#endif
}

/*
 * write additional seglists
 * write object + embedded seglist
 * wait for all segments to be written (async)
 * - body
 * - aux
 *
 * XXX can save some io for inlog == 0
 */

void
fellow_busy_done(struct fellow_busy *fbo, struct objcore *oc, unsigned inlog)
{
	struct fellow_cache_obj *fco;
	struct fellow_disk_obj *fdo;
	struct fellow_cache *fc;

	CHECK_OBJ_NOTNULL(fbo, FELLOW_BUSY_MAGIC);
	fco = fbo->fco;
	CHECK_OBJ_NOTNULL(fco, FELLOW_CACHE_OBJ_MAGIC);
	fdo = FCO_FDO(fco);
	CHECK_OBJ_NOTNULL(fdo, FELLOW_DISK_OBJ_MAGIC);
	fc = fbo->fc;
	CHECK_OBJ_NOTNULL(fc, FELLOW_CACHE_MAGIC);

	// nicer way?
	if (fbo->body_seg != NULL)
		fellow_busy_obj_trimstore(fbo);

	if (fco->seglist.next) {
		AN(fco->seglist.fdsl);
		fellow_cache_seglists_write(fbo,
		    fco->seglist.next, fco->seglist.fdsl->next);
	}

	fellow_disk_seglist_fini(fco->seglist.fdsl);

	AZ(fdo->fdo_flags);
	AZ(fco->oc);
	// XXX still needed with logstate?
	if (inlog) {
		fdo->fdo_flags |= FDO_F_INLOG;
		fco->oc = oc;
	}
	fellow_cache_obj_unbusy(fbo, inlog ? FCOL_WANTLOG : FCOL_NOLOG);
}

static void
fellow_busy_free(struct fellow_busy **fbop)
{
	struct fellow_busy *fbo;
	struct fellow_cache *fc;

	TAKE_OBJ_NOTNULL(fbo, fbop, FELLOW_BUSY_MAGIC);
	fc = fbo->fc;
	CHECK_OBJ_NOTNULL(fc, FELLOW_CACHE_MAGIC);

	buddy_return1_ptr_extent(fc->membuddy, &fbo->fbo_mem);
}

// TODO
// static void v_matchproto_(objslim_f)
// sbu_slim(struct worker *wrk, struct objcore *oc)




/*
 * data is written to the last cache_seglist
 * when done, nsegs is incremented
 *
 * to write segments:
 * - mem alloc space for fdsl
 * - mem alloc space for fcsl & point fcs'es to fds
 * - dsk alloc space for data (disk allocation)
 * - dsk alloc space for fdsl (n x fds)
 *
 * remaining space into first fds
 *
 * disk segment != disk allocation
 * disk segment = checksum, size to write/read
 * disk allocation -> alloc'ed in log
 *
 * rules for writing
 * - always from start to end
 * - if size is known, make one disk allocation
 * - if size is unknown, next alloc must be current + 1
 *
 *
 * the trouble with disk vs. mem segments
 *
 * the log assumes we got max 51 disk segments per obj
 * our memory segments must not be larger than 1/16th - 1/8 of mem
 *
 * so: alloc one big chunk of disk, chop into smaller, individually
 * checksummed segments
 *
 * for chunked: next pow2 with every chunk, so we never get above ~50
 * disk segments
 *
 */

// XXX rearrange seglists ?

static void
fellow_cache_obj_fini(const struct fellow_cache_obj *fco)
{
	struct fellow_disk_seg	*fds;
	struct fellow_disk_obj	*fdo;

	fds = FCO_FCS(fco)->disk_seg;
	CHECK_OBJ_NOTNULL(fds, FELLOW_DISK_SEG_MAGIC);

	fdo = FCO_FDO(fco);
	CHECK_OBJ_NOTNULL(fdo, FELLOW_DISK_OBJ_MAGIC);
	assert(fdo == (void *)FCO_FCS(fco)->alloc.ptr);

	fh(fds->fht, fds->fh, (char *)fdo + FELLOW_DISK_OBJ_CHK_START,
	    FELLOW_DISK_OBJ_CHK_LEN(fdo));
}

static struct fellow_cache_res
fellow_cache_obj_prepread(const struct fellow_cache *fc, fellow_disk_block fdba)
{
	struct fellow_cache_res fcr;
	struct fellow_cache_seg *fcs;
	struct fellow_cache_obj *fco;
	size_t sz;

	sz = fdb_size(fdba);
	if (! PAOK(sz))
		return (FCR_IOFAIL("bad size"));

	fcr = fellow_cache_obj_new(fc, sz, 0);
	if (fcr.status != fcr_ok)
		return (fcr);

	CAST_OBJ_NOTNULL(fco, fcr.r.ptr, FELLOW_CACHE_OBJ_MAGIC);
	fco->fdb = fdba;

	assert(fco->logstate == FCOL_DUNNO);
	fco->logstate = FCOL_INLOG;

	fcs = FCO_FCS(fco);
	CHECK_OBJ(fcs, FELLOW_CACHE_SEG_MAGIC);

	// we prepare the state the object will have
	// when inserted
	fcs->refcnt = 1;
	fcs->fco_infdb = 1;
	fellow_cache_seg_transition_locked_notincore(fcs, FCO_READING);

	return (fcr);
}

// undo fellow_cache_obj_prepread()
static void
fellow_cache_obj_redundant(struct fellow_cache *fc,
    struct fellow_cache_obj **fcop)
{
	struct fellow_cache_obj *fco;
	struct fellow_cache_seg *fcs;

	TAKE_OBJ_NOTNULL(fco, fcop, FELLOW_CACHE_OBJ_MAGIC);
	fcs = FCO_FCS(fco);
	CHECK_OBJ(fcs, FELLOW_CACHE_SEG_MAGIC);

	assert(fcs->refcnt == 1);
	assert(fcs->fco_infdb);
	assert(fcs->state == FCO_READING);

	fcs->refcnt = 0;
	fcs->fco_infdb = 0;
	fellow_cache_seg_transition_locked_notincore(fcs, FCO_REDUNDANT);

	/* holding mutex not necessary here, but this call site is the exception
	 * and the other call side requires it
	 */

	//lint --e{454} locked but not unlocked
	AZ(pthread_mutex_lock(&fco->mtx));
	fellow_cache_obj_free(fc, &fco);
}

/*
 * if returns 0, fellow_cache_obj_free() must be called under the lock
 */
static unsigned
fellow_cache_obj_deref_locked(struct fellow_cache *fc,
    struct fellow_cache_obj *fco)
{
	struct fellow_cache_seg *fcs;
	unsigned refcnt;

	CHECK_OBJ_NOTNULL(fc, FELLOW_CACHE_MAGIC);
	CHECK_OBJ_NOTNULL(fco, FELLOW_CACHE_OBJ_MAGIC);

	fcs = FCO_FCS(fco);
	AN(fcs->refcnt);

	assert_cache_seg_consistency(fcs);

	if (fcs->refcnt == 1 && fcs->fco_infdb) {
		refcnt = fellow_cache_seg_deref_locked(fc, fcs);
		fcs->fco_infdb = 0;
		AZ(refcnt);
		/* REF_FDB_REMOVE */
		AZ(pthread_mutex_lock(&fc->fdb_mtx));
		(void) VRBT_REMOVE(fellow_cache_fdb_head, &fc->fdb_head, fco);
		AZ(pthread_mutex_unlock(&fc->fdb_mtx));
	} else {
		refcnt = fellow_cache_seg_deref_locked(fc, fcs);
	}
	return (refcnt);
}

void
fellow_cache_obj_deref(struct fellow_cache *fc, struct fellow_cache_obj *fco)
{
	unsigned refcount;

	CHECK_OBJ_NOTNULL(fc, FELLOW_CACHE_MAGIC);
	CHECK_OBJ_NOTNULL(fco, FELLOW_CACHE_OBJ_MAGIC);

	AZ(pthread_mutex_lock(&fco->mtx));
	refcount = fellow_cache_obj_deref_locked(fc, fco);
	DBG("fco %p refcount %u", fco, refcount);

	//lint --e{456} Two execution paths are being combined...
	//lint --e{454} A thread mutex has been locked...
	if (refcount == 0)
		fellow_cache_obj_free(fc, &fco);
	else
		AZ(pthread_mutex_unlock(&fco->mtx));
}

/*
 * return with an additional reference
 *
 * if an ocp argument is present and this is the first reference, it gets taken
 *
 */
struct fellow_cache_res
fellow_cache_obj_get(struct fellow_cache *fc,
    struct objcore **ocp, uintptr_t priv2)
{
	struct fellow_cache_res fcr;
	struct fellow_cache_obj *fco, *nfco;
	struct fellow_cache_seg *fcs;
	struct fellow_disk_obj *fdo;
	struct fellow_disk_seglist *fdsl;
	fellow_disk_block fdba;
	unsigned oref = 0;
	const char *err;
	int r;

	CHECK_OBJ_NOTNULL(fc, FELLOW_CACHE_MAGIC);
	fdba.fdb = priv2;

	/*
	 * init a new object, then insert it atomically
	 *
	 * for a race, insert might result in an existing object, in which case
	 * we wasted some work and free the extra object again
	 *
	 */
	fcr =  fellow_cache_obj_prepread(fc, fdba);
	if (fcr.status != fcr_ok) {
		fellow_cache_res_check(fc, fcr);
		return (fcr);
	}
	CAST_OBJ_NOTNULL(nfco, fcr.r.ptr, FELLOW_CACHE_OBJ_MAGIC);

  again:

	//lint --e{456} flexelint does not grok trylock
	AZ(pthread_mutex_lock(&fc->fdb_mtx));
	fco = VRBT_INSERT(fellow_cache_fdb_head, &fc->fdb_head, nfco);
	if (fco != NULL) {
		// the fco might be about to be destroyed see REF_FDB_REMOVE
		r = pthread_mutex_trylock(&fco->mtx);
		if (r != 0) {
			assert(r == EBUSY);
		} else {
			if (FCO_REFCNT(fco))
				oref = ++FCO_REFCNT(fco);
			AZ(pthread_mutex_unlock(&fco->mtx));
		}
	}
	else if (ocp) {
		AZ(nfco->oc);
		TAKE_OBJ_NOTNULL(nfco->oc, ocp, OBJCORE_MAGIC);
	}
	AZ(pthread_mutex_unlock(&fc->fdb_mtx));

	if (fco != NULL && oref == 0) {
		/* we did not manage to grab a reference on the
		 * object
		 */
		goto again;
	}

	if (fco != NULL) {
		// entries on the fdb need at least one reference
		AN(oref);
		// clean up over-allocated new fco
		fellow_cache_obj_redundant(fc, &nfco);

		if (FCO_STATE(fco) == FCO_READING) {
			AZ(pthread_mutex_lock(&fco->mtx));
			while (FCO_STATE(fco) == FCO_READING)
				fellow_cache_seg_wait_locked(FCO_FCS(fco));
			AZ(pthread_mutex_unlock(&fco->mtx));
		}

		switch (FCO_STATE(fco)) {
		case FCO_WRITING:
		case FCO_INCORE:
			break;
		case FCO_READFAIL:
			fellow_cache_obj_deref(fc, fco);
			return (FCR_IOFAIL("hit READFAIL object"));
		case FCO_EVICT:
			// race, retry
			fellow_cache_obj_deref(fc, fco);
			return (fellow_cache_obj_get(fc, ocp, priv2));
		default:
			WRONG("fco state");
		}

		return (FCR_OK(fco));
	}

	TAKEZN(fco, nfco);
	assert(fco->fdb.fdb == fdba.fdb);

	/* not using fellow_cache_seg_read_sync()
	 * because our disk size is smaller than mem size in fcs
	 */
	fcs = &fco->fdo_fcs;

	if (FC_INJ || fellow_io_pread_sync(fc->ffd, fcs->alloc.ptr,
		fdb_size(fdba), fdb_off(fdba)) != (int32_t)fdb_size(fdba)) {
		err = FC_ERRSTR("fdo read error");
		goto err;
	}

	fdo = fellow_disk_obj(fcs);
	AN(fdo);
	err = fellow_disk_obj_check(fdo, fdba);
	if (err != NULL)
		goto err;

	fdsl = (void *)((uint8_t *)fdo + sizeof *fdo + fdo->va_data_len);
	assert(PAOK(fdsl));
	// XXX load of folow-up seglists could be async
	// should not be a common case though, we try to make the first
	// seglist fit
	err = fellow_cache_seglists_load(fc, fco, &fco->seglist, fdsl);
	if (err != NULL)
		goto err;

#define FDO_AUXATTR(U, l)						\
	fellow_cache_seg_associate(&fco->aa_##l##_seg,			\
	    &fdo->aa_##l##_seg, fdo->aa_##l##_seg.seg.size == 0 ?	\
	FCS_USABLE : FCS_DISK);
#include "tbl/fellow_obj_attr.h"

	AZ(pthread_mutex_lock(&fco->mtx));
	assert(fcs->state == FCO_READING);
	assert_cache_seg_consistency(fcs);
	AN(fcs->fco_infdb);
	fcs->disk_seg = &fdo->fdo_fds;
	fellow_cache_seg_transition_locked(fc, fcs,
	    fcs->state, FCO_INCORE);
	if (fcs->refcnt > 1)
		AZ(pthread_cond_broadcast(&fco->cond));
	assert_cache_seg_consistency(fcs);
	AZ(pthread_mutex_unlock(&fco->mtx));

	return (FCR_OK(fco));
  err:
	if (ocp) {
		AZ(*ocp);
		TAKE_OBJ_NOTNULL(*ocp, &fco->oc, OBJCORE_MAGIC);
	}

	/* we are in the process of transitioning to FCO_READING,
	 * fcs is not yet consistent. Just set the failed state,
	 * _evict() will signal any waiters
	 */
	fellow_cache_seg_transition_locked_notincore(FCO_FCS(fco), FCO_READFAIL);
	fellow_cache_obj_deref(fc, fco);

	if (strstr(err, "alloc"))
		fcr = FCR_ALLOCERR(err);
	else
		fcr = FCR_IOERR(err);
	fellow_cache_res_check(fc, fcr);
	return (fcr);
}

static void
fellow_disk_obj_delete_submit(struct fellow_fd *ffd,
    fellow_disk_block start, const uint8_t hash[DIGEST_LEN],
    const struct buddy_off_extent region[FCO_MAX_REGIONS], unsigned n)
{
	unsigned r, ndle = (n / DLE_REG_NREGION) + 2;
	struct fellow_dle *e, dles[ndle];

	fellow_dle_init(dles, ndle);
	if (n)
		r = fellow_dle_reg_fill(dles, ndle, region, n,
		    DLE_REG_DEL_ALLOCED, hash);
	else
		r = 0;

	e = &dles[r++];
	e->type = DLEDSK(DLE_OBJ_DEL_ALLOCED);
	memcpy(e->u.obj.hash, hash, (size_t)DIGEST_LEN);
	e->u.obj.start = start;

	fellow_log_dle_submit(ffd, dles, r);
}

/* evict, then delete from log (this order is important to avoid
 * fdb reuse)
 *
 * will not be called if the object is still in use,
 * so we can delete all segments immediately (via the log)
 *
 * loses one reference
 */
void
fellow_cache_obj_delete(struct fellow_cache *fc,
    struct fellow_cache_obj *fco, const uint8_t hash[DIGEST_LEN])
{
	const struct fellow_disk_obj *fdo;
	/* one additional region for the object itself */
	struct buddy_off_extent region[FCO_MAX_REGIONS + 1] = {{0}};
	struct stvfe_tune *tune;
	enum fcol_state logstate;
	fellow_disk_block fdba;
	unsigned u, n;

	CHECK_OBJ_NOTNULL(fc, FELLOW_CACHE_MAGIC);
	CHECK_OBJ_NOTNULL(fco, FELLOW_CACHE_OBJ_MAGIC);
	fdo = FCO_FDO(fco);
	CHECK_OBJ_NOTNULL(fdo, FELLOW_DISK_OBJ_MAGIC);

	fdba = fco->fdb;

	n = fellow_obj_regions(fc->ffd, fco, region);
	assert(n <= FCO_MAX_REGIONS);

	AZ(pthread_mutex_lock(&fco->mtx));
	/* we must not free the object's disk space while it is still writing */
	while (FCO_STATE(fco) == FCO_WRITING)
		fellow_cache_seg_wait_locked(FCO_FCS(fco));

	switch (fco->logstate) {
	case FCOL_DUNNO:
	case FCOL_WANTLOG:
		fco->logstate = FCOL_TOOLATE;
		break;
	case FCOL_NOLOG:
		break;
	case FCOL_TOOLATE:
		WRONG("fellow_cache_obj_delete FCOL_TOOLATE");
	case FCOL_INLOG:
		fco->logstate = FCOL_DELETED;
		stvfe_oc_log_removed(fco->oc);
		break;
	case FCOL_DELETED:
		WRONG("fellow_cache_obj_delete FCOL_DELETED");
	default:
		WRONG("fellow_cache_obj_delete logstate");
	}

	logstate = fco->logstate;

	//lint --e{456} Two execution paths are being combined...
	//lint --e{454} A thread mutex has been locked...

	u = fellow_cache_obj_deref_locked(fc, fco);
	DBG("fco %p refcount %u", fco, u);
	if (u == 0)
		fellow_cache_obj_free(fc, &fco);
	else
		AZ(pthread_mutex_unlock(&fco->mtx));

	/* UNLOCKED because these states are all final */
	switch (logstate) {
	case FCOL_NOLOG:
	case FCOL_TOOLATE:
		break;
	case FCOL_DELETED:
		fellow_disk_obj_delete_submit(fc->ffd, fdba, hash,
		    region, n);
		return;
	default:
		WRONG("fellow_cache_obj_delete logstate UNLOCKED");
	}

	/* XXX no async IO here yet: We need to free the dskbuddy space after
	 * the discard is complete. To do that, we would need to store the
	 * off/size of the discard in an additional structure of the io ring.
	 *
	 * we can not just hold the cache async io mtx because it would delay
	 * readahead
	 */

	assert(n < FCO_MAX_REGIONS + 1);
	region[n].off = fdb_off(fdba);
	region[n].size = fdb_size(fdba);
	n++;

	tune = fc->tune;
	CHECK_OBJ_NOTNULL(tune, STVFE_TUNE_MAGIC);

	fellow_io_regions_discard(fc->ffd, NULL, region, n,
	    tune->discard_immediate, 1);
	for (u = 0; u < n; u++)
		fellow_region_free(fc->ffd, &region[u]);
}

// XXX status
void *
fellow_cache_obj_getattr(struct fellow_cache *fc,
    struct fellow_cache_obj *fco,
    enum obj_attr attr, size_t *len)
{
	struct fellow_disk_obj *fdo;

	CHECK_OBJ_NOTNULL(fco, FELLOW_CACHE_OBJ_MAGIC);
	fdo = FCO_FDO(fco);
	CHECK_OBJ_NOTNULL(fdo, FELLOW_DISK_OBJ_MAGIC);
	AN(len);

	switch (attr) {
		/* Fixed size attributes */
#define FDO_FIXATTR(U, l, s, vs)					\
	case OA_##U:							\
		*len = vs;						\
		return (fdo->fa_##l);
#include "tbl/fellow_obj_attr.h"

		/* Variable size attributes */
#define FDO_VARATTR(U, l)						\
	case OA_##U:							\
		if (fdo->va_##l.aoff == 0)				\
			return (NULL);					\
		*len = fdo->va_##l.alen;				\
		return ((uint8_t *)fdo + fdo->va_##l.aoff);
#include "tbl/fellow_obj_attr.h"

		/* Auxiliary attributes
		 * ref is never returned and just ignored by
		 * fellow_cache_seg_auxattr_free()
		 */
#define FDO_AUXATTR(U, l)						\
	case OA_##U: {							\
		struct fellow_cache_seg *fcs = &fco->aa_##l##_seg;	\
		if (fcs->disk_seg->seg.size == 0)			\
			return (NULL);					\
		fellow_cache_seg_ref_in(fc, FCIO_SYNC, &fcs, 1);	\
		XXXAZ(fellow_cache_seg_check(fc, fcs));			\
		*len = fcs->len;					\
		return (fcs->alloc.ptr);				\
	}
#include "tbl/fellow_obj_attr.h"

	default:
		break;
	}
	WRONG("Unsupported OBJ_ATTR");
}

void *
fellow_busy_setattr(struct fellow_busy *fbo, enum obj_attr attr,
    size_t len, const void *ptr)
{
	struct fellow_cache_obj *fco;
	struct fellow_disk_obj *fdo;
	struct fellow_cache *fc;
	void *retval = NULL;
	size_t sz;
	ssize_t diff;
	uint8_t *p;

	CHECK_OBJ_NOTNULL(fbo, FELLOW_BUSY_MAGIC);
	fc = fbo->fc;
	CHECK_OBJ_NOTNULL(fc, FELLOW_CACHE_MAGIC);
	fco = fbo->fco;
	CHECK_OBJ_NOTNULL(fco, FELLOW_CACHE_OBJ_MAGIC);
	fdo = FCO_FDO(fco);
	CHECK_OBJ_NOTNULL(fdo, FELLOW_DISK_OBJ_MAGIC);

	assert(fbo->va_data);
	assert(fbo->va_data >= fdo->va_data);
	assert(fbo->va_data <= fdo->va_data + fdo->va_data_len);
	assert(fbo->va_data_len <= fdo->va_data_len);

	switch (attr) {
		/* Fixed size attributes */
#define FDO_FIXATTR(U, l, s, vs)					\
	case OA_##U:							\
		assert(len <= sizeof fdo->fa_##l);			\
		retval = fdo->fa_##l;					\
		break;
#include "tbl/fellow_obj_attr.h"

		/* Variable size attributes */
#define FDO_VARATTR(U, l)						\
	case OA_##U:							\
		p = (uint8_t *)fdo;					\
		if (fdo->va_##l.alen > 0) {				\
			AN(fdo->va_##l.aoff);				\
			assert(len == fdo->va_##l.alen);		\
			retval = p + fdo->va_##l.aoff;			\
		} else if (len > 0) {					\
			assert(len <= UINT32_MAX);			\
			sz = fbo->va_data_len + len;			\
			assert(sz <= fdo->va_data_len);			\
			retval = fbo->va_data;				\
			diff = fbo->va_data - p;			\
			assert(diff > 0);				\
			assert(diff < UINT32_MAX);			\
			fdo->va_##l.aoff = (uint32_t)diff;		\
			fdo->va_##l.alen = (uint32_t)len;		\
			fbo->va_data += len;				\
		}							\
		break;
#include "tbl/fellow_obj_attr.h"

		/* Auxiliary attributes

		 * XXX: We only support writing once, this
		 * is different to varnish-cache
		 */
#define FDO_AUXATTR(U, l)						\
	case OA_##U: {							\
		struct fellow_cache_seg *fcs = &fco->aa_##l##_seg;	\
		if (len == 0)						\
			break;						\
		if (! fellow_busy_seg_alloc(fbo, fcs, len))		\
			break;						\
		assert(fcs->alloc.size >= len);				\
		AZ(fcs->len);						\
		fcs->len = len;						\
		retval = fcs->alloc.ptr;				\
		memcpy(retval, ptr, len);				\
		fellow_cache_seg_unbusy(fbo, fcs);			\
		return (retval);					\
	}
#include "tbl/fellow_obj_attr.h"

	default:
		WRONG("Unsupported OBJ_ATTR");
		break;
	}

	// !! early return for AUXATTR above !
	if (retval != NULL && ptr != NULL)
		memcpy(retval, ptr, len);
	return (retval);
}

static void
fellow_cache_assert_disk_fmt(void)
{
	//lint --e{506} constant value boolean

	assert(sizeof(uintptr_t) == sizeof (fellow_disk_block));

#define S struct fellow_disk_seg
	assert(offsetof(S, segnum) == 4);
	assert(offsetof(S, seg) == 8);
	assert(offsetof(S, fh) == 24);
	assert(sizeof(S) == 24 + 32);
	assert(sizeof(S) == 7 * 8);
#undef S

#define S struct fellow_disk_seglist
	assert(offsetof(S, version) == 4);
	assert(offsetof(S, fht) == 7);
	assert(offsetof(S, fh) == 8);
	assert(offsetof(S, next) == 40);
	assert(offsetof(S, nsegs) == 56);
	assert(offsetof(S, lsegs) == 58);
	assert(offsetof(S, segs) == 64);
	assert(sizeof(S) == 64);
#undef S

#define S struct fellow_disk_obj
	assert(offsetof(S, version) == 4);
	assert(offsetof(S, fdo_fds) == 8);
	assert(offsetof(S, va_data_len) == 64);

	assert(offsetof(S, fa_flags) == 70);
	assert(offsetof(S, fa_len) == 9 * 8);
	assert(offsetof(S, fa_vxid) == 10 * 8);
	assert(offsetof(S, fa_lastmodified) == 11 * 8);
	assert(offsetof(S, fa_gzipbits) == 12 * 8);
	assert(offsetof(S, fa_reserve) == 128);

	assert(offsetof(S, va_vary) == 192);
	assert(offsetof(S, va_headers) == 200);
	assert(offsetof(S, va_reserve) == 208);

	assert(offsetof(S, aa_esidata_seg) == 240);
	assert(offsetof(S, va_data) == 296);
	assert(offsetof(S, va_data) == 37 * 8);
#undef S

}

struct fellow_cache *
fellow_cache_init(struct fellow_fd *ffd, buddy_t *membuddy,
    struct stvfe_tune *tune, fellow_task_run_t taskrun)
{
	struct fellow_cache *fc;

	AN(ffd);
	AN(membuddy);

	fellow_cache_assert_disk_fmt();

	ALLOC_OBJ(fc, FELLOW_CACHE_MAGIC);
	AN(fc);

	fc->ffd = ffd;
	fc->membuddy = membuddy;
	fc->tune = tune;
	fc->running = 1;

	AZ(pthread_mutex_init(&fc->fdb_mtx, NULL));
	AZ(pthread_mutex_init(&fc->lru_mtx, NULL));

	VRBT_INIT(&fc->fdb_head);
	VTAILQ_INIT(&fc->lru_head);

	fellow_cache_async_init(fc, taskrun);

	return (fc);
}

/*
 * the LRU thread is started by the storage layer after loading
 */
void
fellow_cache_fini(struct fellow_cache **fcp, const pthread_t *lru_thread)
{
	struct fellow_cache *fc;
	void *r;

	TAKE_OBJ_NOTNULL(fc, fcp, FELLOW_CACHE_MAGIC);

	fc->running = 0;
	buddy_wait_kick(fc->membuddy);
	if (lru_thread != NULL) {
		AZ(pthread_join(*lru_thread, &r));
		AZ(r);
	}

	assert(VTAILQ_EMPTY(&fc->lru_head));
	assert(VRBT_EMPTY(&fc->fdb_head));

	fellow_cache_async_fini(fc);

	AZ(pthread_mutex_destroy(&fc->lru_mtx));
	AZ(pthread_mutex_destroy(&fc->fdb_mtx));

	FREE_OBJ(fc);
}

#ifdef TEST_DRIVER
#include "vsha256.h"
#include "fellow_testenv.h"

#ifdef DEBUG
const char * const filename = "/tmp/fellowfile.cachetest";
#else
const char * const filename = "/tmp/fellowfile.cachetest.ndebug";
#endif

static int v_matchproto_(fellow_resurrect_f)
resurrect_keep(void *priv, const struct fellow_dle *e)
{

	(void) priv;
	(void) e;

	return (1);
}

static int v_matchproto_(fellow_resurrect_f)
resurrect_discard(void *priv, const struct fellow_dle *e)
{

	(void) priv;
	(void) e;

	return (0);
}

static int
iter_sha256(void *priv, unsigned flush, const void *ptr, ssize_t l)
{
	VSHA256_CTX *sha256ctx = priv;

	(void) flush;
	VSHA256_Update(sha256ctx, ptr, l);
	return (0);
}

static const char * const oatest[OA__MAX] = {
	[OA_LEN] = "lenlenle",
	[OA_VXID] = "vxid",
	[OA_FLAGS] = "f",
	[OA_GZIPBITS] = "gzipbitsgzipbitsgzipbitsgzipbits",
	[OA_LASTMODIFIED] = "lastmodi",
	[OA_VARY] = "vary vary vary varyvaryvary vary varyvaryvary",
	[OA_HEADERS] = "headers hdrs hea",
	[OA_ESIDATA] = "This is OA_ESIDATA <EOS>"
};

void t_getattr(struct fellow_cache *fc, struct fellow_cache_obj *fco)
{
	unsigned u;
	size_t sz;
	const char *v;

	for (u = 0; u < OA__MAX; u++) {
		v = fellow_cache_obj_getattr(fc, fco, u, &sz);
		AN(v);
		assert(sz == strlen(oatest[u]));
		AZ(strncmp(v, oatest[u], sz));
	}
}

/* mimic sfemem_bocdone() without fellow_storage */

int
stvfe_oc_inlog(struct objcore *oc)
{
	(void)oc;
	return (1);
}

void
stvfe_oc_log_removed(struct objcore *oc)
{
	(void)oc;
	return;
}

void
stvfe_oc_log_submitted(struct objcore *oc)
{
	(void)oc;
	return;
}

void
stvfe_oc_dle_obj(struct objcore *oc, struct fellow_dle *e)
{
	AN(oc);	// actually the hash
	AN(e);
	memcpy(e->u.obj.hash, oc, DIGEST_LEN);
	e->u.obj.ttl = 1;
}

static void
test_bocdone(struct fellow_busy *fbo, const uint8_t *hash, unsigned incore)
{
	fellow_busy_done(fbo, TRUST_ME(hash), incore);
}

enum incache_e {
	notincache = 0,
	incache = 1
};

/*
 * ======================================================================
 * test fellow cache object creation / unbusy with error injection
 * during write
 */
static void
test_fellow_cache_unbusy_inject(struct fellow_cache *fc)
{
	struct fellow_cache_obj *fco;
	struct fellow_cache_res fcr;
	struct fellow_busy *fbo;
	uint8_t hash[DIGEST_LEN];
	uintptr_t priv2;
	int injcount = -1;
	size_t sz = 1234;
	char *ptr;

	fc_inj_reset();
	while (injcount) {
		DBG("injcount=%d", injcount);
		if (injcount > 0)
			fc_inj_set(injcount);
		fcr = fellow_busy_obj_alloc(fc, &fco, &priv2, 1234);
		if (fcr.status != fcr_ok) {
			DBG("fbo_alloc %d err %s", injcount, fcr.r.err);
			goto next;
		}

		fbo = fcr.r.ptr;
		CHECK_OBJ_NOTNULL(fbo, FELLOW_BUSY_MAGIC);

		fcr = fellow_busy_obj_getspace(fbo, &sz, (uint8_t **)&ptr);

		if (fcr.status == fcr_ok)
			fellow_busy_obj_extend(fbo, sz);
		else
			DBG("getspace inj %d err %s", injcount, fcr.r.err);

		fellow_busy_obj_trimstore(fbo);

		test_bocdone(fbo, TRUST_ME(hash), 1);
		fellow_cache_obj_deref(fc, fco);

	  next:
		if (injcount < 0) {
			injcount = fc_inj_count();
			/*
			 * XXX TODO
			 HIT 6 = fellow_cache_async_write_complete:2171
			 Incomplete code in fellow_cache_async_write_complete(), fellow_cache.c line 2175:
			 Aborted
			 *
			 */
			injcount -= 2;	// XXX
		} else {
			// XXX check obj
			injcount--;
		}
	}
	fc_inj_reset();
}

/*
 * ======================================================================
 * test fellow cache with error injection at every possible point
 */

static struct fellow_cache_obj *
test_fellow_cache_obj_get(struct fellow_cache *fc, uintptr_t priv2,
    enum incache_e expect)
{
	struct fellow_cache_res fcr;
	struct fellow_cache_obj *fco;
	struct objcore *oc1, *toc, ocmem[1];
	int injcount;

	INIT_OBJ(ocmem, OBJCORE_MAGIC);

	DBG("expect %s", expect == incache ? "incache" : "notincache");

	// baseline, no injection
	fc_inj_reset();
	toc = ocmem;
	fcr = fellow_cache_obj_get(fc, &toc, priv2);
	if (fcr.status != fcr_ok) {
		AN(toc);
		return (NULL);
	}
	CAST_OBJ_NOTNULL(fco, fcr.r.ptr, FELLOW_CACHE_OBJ_MAGIC);
	DBG("fco1 %p ref %u", fco, FCO_REFCNT(fco));
	if (expect == incache)
		AN(toc);
	else
		AZ(toc);
	oc1 = toc;

	injcount = fc_inj_count();
	DBG("injcount %d", injcount);
	if (injcount == 0)
		return (fco);

	fellow_cache_obj_deref(fc, fco);

	AN(injcount);
	while (injcount) {
		fc_inj_set(injcount);
		toc = ocmem;
		fcr = fellow_cache_obj_get(fc, &toc, priv2);
		AN(toc);
		assert(fcr.status != fcr_ok);
		DBG("inj %d err %s", injcount, fcr.r.err);
		injcount--;
	}

	DBG("fco1e %p ref %u", fco, FCO_REFCNT(fco));

	fc_inj_reset();
	toc = ocmem;
	fcr = fellow_cache_obj_get(fc, &toc, priv2);
	DBG("fco2 %p ref %u", fco, FCO_REFCNT(fco));
	assert(toc == oc1);
	CAST_OBJ_NOTNULL(fco, fcr.r.ptr, FELLOW_CACHE_OBJ_MAGIC);
	return (fco);
}

static void test_fellow_cache_obj_iter_final(
    struct fellow_cache *fc, struct fellow_cache_obj **fcop,
    const unsigned char h1[SHA256_LEN], int final)
{
	struct fellow_cache_res fcr;
	struct fellow_cache_obj *fco;
	unsigned char h2[SHA256_LEN];
	struct objcore *toc, ocmem[1];
	VSHA256_CTX sha256ctx;
	uintptr_t priv2;
	unsigned u, refcnt;
	int injcount;

	INIT_OBJ(ocmem, OBJCORE_MAGIC);

	AN(fcop);
	AN(*fcop);

	// baseline, no injection
	fc_inj_reset();
	VSHA256_Init(&sha256ctx);
	fcr = fellow_cache_obj_iter(fc, *fcop, &sha256ctx, iter_sha256, final);
	assert(fcr.status == fcr_ok);
	AZ(fcr.r.integer);
	VSHA256_Final(h2, &sha256ctx);
	AZ(memcmp(h1, h2, sizeof *h1));

	injcount = fc_inj_count();
	DBG("injcount %d", injcount);
	if (injcount == 0)
		return;

	// remember object and deref
	priv2 = (*fcop)->fdb.fdb;
	refcnt = FCO_REFCNT(*fcop);
	for (u = 0; u < refcnt; u++)
		(void) fellow_cache_obj_deref(fc, *fcop);
	*fcop = NULL;

	AN(injcount);
	while (injcount) {
		toc = ocmem;
		fcr = fellow_cache_obj_get(fc, &toc, priv2);
		assert(fcr.status == fcr_ok);
		fco = fcr.r.ptr;

		fc_inj_set(injcount);
		VSHA256_Init(&sha256ctx);
		fcr = fellow_cache_obj_iter(fc, fco, &sha256ctx,
		    iter_sha256, 0);
		assert (fcr.status != fcr_ok);
		DBG("inj %d err %s", injcount, fcr.r.err);
		injcount--;

		assert(FCO_REFCNT(fco) == 1);
		fellow_cache_obj_deref(fc, fco);
	}

	// get back the original number of references
	for (u = 0; u < refcnt; u++) {
		toc = ocmem;
		fcr = fellow_cache_obj_get(fc, &toc, priv2);
		assert(fcr.status == fcr_ok);
		assert((toc == NULL) == (u == 0));
		fco = fcr.r.ptr;
	}
	*fcop = fco;
}

static void test_fellow_cache_obj_iter(
    struct fellow_cache *fc, struct fellow_cache_obj **fcop,
    const unsigned char h1[SHA256_LEN])
{
	int final;

	AN(fcop);
	AN(*fcop);

	for (final = 0; final < 2; final++)
		test_fellow_cache_obj_iter_final(fc, fcop, h1, final);
}

#if 0
void *fellow_cache_obj_getattr(struct fellow_cache *fc,
    struct fellow_cache_obj *fco,
    enum obj_attr attr, size_t *len);
int fellow_cache_obj_lru_touch(struct fellow_cache *fc,
    struct fellow_cache_obj *fco);

struct fellow_cache_res
fellow_busy_obj_alloc(struct fellow_cache *fc,
    struct fellow_cache_obj **fcop, uintptr_t *priv2,
    unsigned wsl);
struct fellow_cache_res
fellow_busy_obj_getspace(struct fellow_busy *fbo,
    size_t *sz, uint8_t **ptr);
void fellow_busy_obj_extend(struct fellow_busy *fbo, size_t l);
void fellow_busy_obj_trimstore(struct fellow_busy *fbo);
#endif

#define DBGSZ(x) \
	DBG(#x "\t%3zu", sizeof(struct x))

void
t_cache(unsigned chksum)
{
	size_t sz;
	struct fellow_fd *ffd;
	buddy_t mb, *membuddy = &mb;
	struct fellow_cache *fc;
	struct fellow_cache_obj *fco;
	struct fellow_busy *fbo;
	uintptr_t priv2;
	void *resur_priv = NULL;
	unsigned u;
	uint8_t hash[DIGEST_LEN];
	VSHA256_CTX sha256ctx;
	unsigned char h1[SHA256_LEN];
	struct stvfe_tune tune[1];
	const size_t memsz = 10 * 1024 * 1024;
	const size_t dsksz = 100 * 1024 * 1024;
	const size_t objsize_hint = 1 * 1024 * 1024;

	DBGSZ(fellow_disk_seg);
	DBGSZ(fellow_disk_seglist);
	DBGSZ(fellow_disk_obj);

	DBGSZ(fellow_cache_seg);
	DBGSZ(fellow_cache_seglist);
	DBGSZ(fellow_cache_obj);

	// canary so size increase does not happen unnoticed
	sz  = sizeof(struct fellow_cache_obj);
	assert(sz <= 320);

	AZ(stvfe_tune_init(tune, memsz, dsksz, objsize_hint));
	tune->hash_obj = chksum;
	// for injection tests
	tune->ioerr_obj = 1;
	tune->ioerr_log = 1;
	tune->allocerr_obj = 1;
	tune->allocerr_log = 1;
	AZ(stvfe_tune_check(tune));

	buddy_init(membuddy, MIN_BUDDY_BITS, memsz,
	    BUDDYF(mmap), NULL, NULL, NULL);
	ffd = fellow_log_init(filename, dsksz, objsize_hint,
	    NULL, NULL, membuddy, tune, fellow_simple_task_run);
	XXXAN(ffd);
	fellow_log_open(ffd, resurrect_discard, &resur_priv);

	fc = fellow_cache_init(ffd, membuddy, tune, fellow_simple_task_run);

	// === empty obj
	fbo = fellow_busy_obj_alloc(fc, &fco, &priv2, 1234).r.ptr;
	CHECK_OBJ_NOTNULL(fbo, FELLOW_BUSY_MAGIC);

	test_bocdone(fbo, TRUST_ME(hash), 1);
	fellow_cache_obj_deref(fc, fco);

	// === alloc space, dont use
	fbo = fellow_busy_obj_alloc(fc, &fco, &priv2, 1234).r.ptr;
	CHECK_OBJ_NOTNULL(fbo, FELLOW_BUSY_MAGIC);
	{
		char *ptr;
		sz = 1234;
		AZ(fellow_busy_obj_getspace(fbo, &sz, (uint8_t **)&ptr).status);
	}
	test_bocdone(fbo, TRUST_ME(hash), 1);
	fellow_cache_obj_deref(fc, fco);

	// === alloc space, trim
	fbo = fellow_busy_obj_alloc(fc, &fco, &priv2, 1234).r.ptr;
	CHECK_OBJ_NOTNULL(fbo, FELLOW_BUSY_MAGIC);
	{
		char *ptr;
		sz = 1234;
		AZ(fellow_busy_obj_getspace(fbo, &sz, (uint8_t **)&ptr).status);
	}
	fellow_busy_obj_trimstore(fbo);
	test_bocdone(fbo, TRUST_ME(hash), 1);
	fellow_cache_obj_deref(fc, fco);

	test_fellow_cache_unbusy_inject(fc);

	// === alloc, then evict
	VSHA256_Init(&sha256ctx);
	fbo = fellow_busy_obj_alloc(fc, &fco, &priv2, 1234).r.ptr;
	CHECK_OBJ_NOTNULL(fbo, FELLOW_BUSY_MAGIC);
	u = fco->seglist.fdsl->lsegs + 72 * 2;
	while (u-- > 0) {
		void *p;
		char *ptr;
		sz = 1234;

		AZ(fellow_busy_obj_getspace(fbo, &sz, (uint8_t **)&ptr).status);
		memset(ptr, 0, sz);
		p = ptr;
		(void) snprintf(ptr, sz, "We got %zu bytes at %p", sz, p);
		if (sz > 1234)
			sz = 1234;
		VSHA256_Update(&sha256ctx, ptr, sz);
		fellow_busy_obj_extend(fbo, sz);
	}
	VSHA256_Final(h1, &sha256ctx);
	fellow_busy_obj_trimstore(fbo);
	// if this fails, u is too low
	AN(fco->seglist.next);
	// fixattr always return a pointer
	for (u = OA_VARY; u < OA__MAX; u++)
		AZ(fellow_cache_obj_getattr(fc, fco, u, &sz));
	for (u = 0; u < OA__MAX; u++)
		AN(fellow_busy_setattr(fbo, u, strlen(oatest[u]), oatest[u]));
	t_getattr(fc, fco);
	test_bocdone(fbo, TRUST_ME(hash), 1);
	priv2 = fco->fdb.fdb;

	/* dumb wait until writes are complete */
	while (FCO_REFCNT(fco) > 1)
		usleep(1000);
	fellow_cache_obj_deref(fc, fco);

	/* get back object without and with close/open */
	for (u = 0; u < 2; u++) {
		fco = test_fellow_cache_obj_get(fc, priv2, notincache);
		assert_cache_seg_consistency(FCO_FCS(fco));
		DBG("fco1 %p ref %u", fco, FCO_REFCNT(fco));

		// second access does not assign the oc
		AN(fco == test_fellow_cache_obj_get(fc, priv2, incache));
		assert_cache_seg_consistency(FCO_FCS(fco));
		DBG("fco2 %p ref %u", fco, FCO_REFCNT(fco));

		// get attrs and compare
		t_getattr(fc, fco);

		// get segments and compare
		test_fellow_cache_obj_iter(fc, &fco, h1);

		assert_cache_seg_consistency(FCO_FCS(fco));

		fellow_cache_obj_deref(fc, fco);

		DBG("fco3 %p ref %u", fco, FCO_REFCNT(fco));
		if (u == 0)
			fellow_cache_obj_deref(fc, fco);
		else
			fellow_cache_obj_delete(fc, fco, hash);


		fellow_cache_fini(&fc, NULL);
		AZ(fc);
		fellow_log_close(&ffd);
		BWIT_ISEMPTY(membuddy->witness);
		buddy_fini(&membuddy, BUDDYF(unmap), NULL, NULL, NULL);
		AZ(ffd);

		if (u == 1)
			break;

		membuddy = &mb;
		buddy_init(membuddy, MIN_BUDDY_BITS, 10 * 1024 * 1024,
		    BUDDYF(mmap), NULL, NULL, NULL);
		ffd = fellow_log_init(filename, dsksz, objsize_hint,
		    NULL, NULL, membuddy, tune, fellow_simple_task_run);
		XXXAN(ffd);
		fellow_log_open(ffd, resurrect_keep, &resur_priv);

		fc = fellow_cache_init(ffd, membuddy, tune,
		    fellow_simple_task_run);
	}
}

int
main(int argc, char **argv)
{
	unsigned l, hash;

#ifdef HAVE_XXHASH_H
	l = FH_LIM;
#else
	l = FH_SHA256 + 1;
#endif

	for (hash = FH_SHA256; hash < l; hash++)
		t_cache(hash);
	printf("OK\n");
	return (0);
}
#endif
