/*-
 * SPDX-License-Identifier: LGPL-2.1-only
 * Copyright 2022 UPLEX Nils Goroll Systemoptimierung. All rights reserved.
 * Author: Nils Goroll <nils.goroll@uplex.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301 USA Also add information on how to contact you by
 * electronic and paper mail.
 */

#include "config.h"

#include <assert.h>
#include <string.h>

#ifdef HAVE_XXHASH_H
#define MIN_XXH_VERSION 800
#include <xxhash.h>
#if XXH_VERSION_NUMBER < MIN_XXH_VERSION
#error "xxhash version unsupported"
#endif
static void __attribute__((constructor))
fh_assert_xxhash_version(void)
{
	assert(XXH_versionNumber() >= MIN_XXH_VERSION);
}
#undef MIN_XXH_VERSION
#endif

#include "fellow_sha256.h"
#include "fellow_hash.h"

static const size_t fh_len[FH_LIM] = {
	[FH_NONE]   = 0,
	[FH_SHA256] = FSHA256_LEN,
#ifdef HAVE_XXHASH_H
	[FH_XXH32]  = sizeof(XXH32_hash_t),
	[FH_XXH3_64]  = sizeof(XXH64_hash_t),
	[FH_XXH3_128] = sizeof(XXH128_hash_t)
#endif
};

const char * const fh_name[FH_LIM] = {
	[FH_NONE]   = "none",
	[FH_SHA256] = "sha256",
#ifdef HAVE_XXHASH_H
	[FH_XXH32]    = "xxh32",
	[FH_XXH3_64]  = "xxh3_64",
	[FH_XXH3_128] = "xxh3_128"
#endif
};

static void __attribute__((constructor))
fh_assert_lengths(void)
{
	//lint -e{708}
	union fh fhh = {{0}};

	(void)fhh;
	assert(fh_len[FH_SHA256] == sizeof fhh.sha256);
#ifdef HAVE_XXHASH_H
	assert(fh_len[FH_XXH32]  == sizeof fhh.xxh32);
	assert(fh_len[FH_XXH3_64]  == sizeof fhh.xxh64);
	assert(fh_len[FH_XXH3_128] == sizeof fhh.xxh128);
#endif
}

void
fh(uint8_t fht, union fh *fhh, const void *p, size_t l)
{
	//lint --e{779,506} == NULL comparison with constant
	switch (fht) {
	case FH_SHA256:
		sha256(fhh->sha256, p, l);
		break;
#ifdef HAVE_XXHASH_H
	case FH_XXH32:
		fhh->xxh32 = XXH32(p, l, 0);
		break;
	case FH_XXH3_64:
		fhh->xxh64 = XXH3_64bits(p, l);
		break;
	case FH_XXH3_128: {
		XXH128_hash_t h = XXH3_128bits(p, l);
		memcpy(fhh->xxh128, &h, sizeof fhh->xxh128);
		break;
	}
#else
	case FH_XXH32:
	case FH_XXH3_64:
	case FH_XXH3_128:
		assert("xxhash support not compiled in" == NULL);
#endif
	default:
		assert("wrong hash type" == NULL);
	}
}

int
fhcmp(uint8_t fht, const union fh *fhh, const void *p, size_t l)
{
	//lint --e{779,506} == NULL comparison with constant
	switch (fht) {
	case FH_SHA256:
		return (sha256cmp(fhh->sha256, p, l));
#ifdef HAVE_XXHASH_H
	case FH_XXH32:
		return (fhh->xxh32 != XXH32(p, l, 0));
	case FH_XXH3_64:
		return (fhh->xxh64 != XXH3_64bits(p, l));
	case FH_XXH3_128:
	{
		XXH128_hash_t h = XXH3_128bits(p, l);
		return (memcmp(fhh->xxh128, &h, sizeof fhh->xxh128));
	}
#else
	case FH_XXH32:
	case FH_XXH3_64:
	case FH_XXH3_128:
		assert("xxhash support not compiled in" == NULL);
#endif
	default:
		assert("wrong hash type" == NULL);
	}
	// SunOS gcc says "control reaches...", flint says not
	//lint -e{527}
	return (-1);
}
