/*-
 * SPDX-License-Identifier: LGPL-2.1-only
 * Copyright 2022,2023 UPLEX Nils Goroll Systemoptimierung. All rights reserved.
 * Author: Nils Goroll <nils.goroll@uplex.de>
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public License
 * as published by the Free Software Foundation; either version 2.1 of
 * the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
 * 02110-1301 USA Also add information on how to contact you by
 * electronic and paper mail.
 */

#ifdef HAVE_LINUX_FS_H
#include <linux/fs.h>
#include <sys/ioctl.h>
#endif
#ifdef HAVE_FALLOCATE
//lint -e(537)
#include <fcntl.h>
#endif

/*
 * list of regions to free:
 *
 * while we iterate on-disk logs, we can not free extra log blocks immediately
 * and while we flush and rewrite logs, we can not free data blocks immediately,
 * they represent our current log and data which is still referenced by the
 * previous log.
 *
 * We need to finish connecting the new log (LBUF_CAN_REF) before we can free
 * them so make sure that when the log is read, it finds the data which was
 * there before the free of the region.
 *
 */
struct regl {
	unsigned			magic;
#define REGL_MAGIC			0xe946b3c2
	unsigned			n, space;
	struct buddy_ptr_page		alloc;
	VSTAILQ_ENTRY(regl)		list;
	struct buddy_off_extent	arr[];
};

struct regionlist {
	unsigned			magic;
#define REGIONLIST_MAGIC		0xeb869815
	VSTAILQ_HEAD(,regl)		head;
	buddy_t				*membuddy;
};

// usage: struct regl_stk(42)
#define regionlist_stk(size)					\
	{							\
		unsigned n, space;				\
		struct regionlist *regionlist;			\
		struct buddy_off_extent	arr[size];	\
	}

#define regionlist_stk_init(size, rl) {			\
		.n = 0,						\
		.space = (size),				\
		.regionlist = (rl),				\
		.arr = {{0}}					\
	}

//lint -e773
#define regionlist_stk_decl(name, size, rl)				\
	struct regionlist_stk(size) name = regionlist_stk_init(size, rl)
//lint +e773

#define regionlist_stk_flush(name)					\
do {									\
	if (name.n > 0) {						\
		regionlist_add(name.regionlist, name.arr, name.n);	\
	}								\
	name.n = 0;							\
} while(0)

#define regionlist_stk_add(name, o, s)					\
do {									\
	if (name.n == name.space) {					\
		regionlist_stk_flush(name);				\
	}								\
	assert(name.n < name.space);					\
	name.arr[name.n].off = (o);					\
	name.arr[name.n].size = (s);					\
	name.n++;							\
} while(0)

static struct regl *
regl_alloc_bits(buddy_t *membuddy, const uint8_t bits, int8_t cram,
    struct regionlist **rlp)
{
	struct buddy_ptr_page alloc;
	struct regionlist *rl;
	struct regl *r;
	int8_t c;
	size_t sz, b;

	sz = 2 * sizeof *r;
	b = log2up(sz);
	assert(bits >= b);

	// should we also support negative cram?
	assert(cram >= 0);
	assert(bits < INT8_MAX);
	c = (int8_t)(bits - (typeof(bits))b);
	assert(c >= 0);
	if (cram > c)
		cram = c;

	alloc = buddy_alloc1_ptr_page_wait(membuddy, FEP_FREE, bits, cram);
	AN(alloc.ptr);

	sz = (size_t)1 << alloc.bits;
	assert(sz >= (sizeof *r + sizeof *rl));

	memset(alloc.ptr, 0, sz);

	if (rlp) {
		rl = alloc.ptr;
		rl->magic = REGIONLIST_MAGIC;
		VSTAILQ_INIT(&rl->head);
		rl->membuddy = membuddy;
		sz -= sizeof(*rl);
		*rlp = rl;

		r = (void *)(rl + 1);
	}
	else {
		r = alloc.ptr;
	}

	sz -= sizeof(*r);

	r->magic = REGL_MAGIC;
	b = sz / sizeof *r->arr;
	assert(b <= UINT_MAX);
	r->space = (unsigned)b;
	r->alloc = alloc;
	return (r);
}

#ifdef DEBUG
static const int8_t regl_bits = 12;
#else
static const int8_t regl_bits = 16;	// XXX 64k reasonable?
#endif

static struct regl *
regl_alloc(buddy_t *membuddy)
{
	return (regl_alloc_bits(membuddy, regl_bits, regl_bits, NULL));
}

static struct regionlist *
regionlist_alloc(buddy_t *membuddy)
{
	struct regionlist *rl;
	struct regl *r;

	r = regl_alloc_bits(membuddy, regl_bits, regl_bits, &rl);
	AN(r);
	AN(rl);

	VSTAILQ_INSERT_HEAD(&rl->head, r, list);
	return (rl);
}

static void
regionlist_append(struct regionlist *to, struct regionlist **fromp)
{
	struct regionlist *from;

	CHECK_OBJ_NOTNULL(to, REGIONLIST_MAGIC);
	TAKE_OBJ_NOTNULL(from, fromp, REGIONLIST_MAGIC);

	VSTAILQ_CONCAT(&to->head, &from->head);
}

static struct regl *
regionlist_extend(struct regionlist *rl)
{
	struct regl *regl;

	CHECK_OBJ_NOTNULL(rl, REGIONLIST_MAGIC);

	regl = regl_alloc(rl->membuddy);
	VSTAILQ_INSERT_TAIL(&rl->head, regl, list);
	return (regl);
}

static void
regionlist_add(struct regionlist *rl,
    const struct buddy_off_extent *regs, unsigned n)
{
	struct regl *regl;
	unsigned av, nn;

	regl = VSTAILQ_LAST(&rl->head, regl, list);
	CHECK_OBJ_NOTNULL(regl, REGL_MAGIC);

	while (n) {
		if (regl->n == regl->space)
			regl = regionlist_extend(rl);

		CHECK_OBJ_NOTNULL(regl, REGL_MAGIC);
		assert(regl->space > regl->n);

		av = regl->space - regl->n;

		if (av < n)
			nn = av;
		else
			nn = n;

		memcpy(&regl->arr[regl->n], regs, nn * sizeof *regs);
		regl->n += nn;
		regs += nn;
		n -= nn;
	}
}

static void
regionlist_free(struct regionlist **rlp, buddy_t *dskbuddy)
{
	struct buddy_returns *dskret, *memret;
	struct regionlist *rl;
	struct regl *regl, *save;
	unsigned n;

	TAKE_OBJ_NOTNULL(rl, rlp, REGIONLIST_MAGIC);

	dskret = BUDDY_RETURNS_STK(dskbuddy, BUDDY_RETURNS_MAX);
	memret = BUDDY_RETURNS_STK(rl->membuddy, BUDDY_RETURNS_MAX);

	VSTAILQ_FOREACH_SAFE(regl, &rl->head, list, save) {
		CHECK_OBJ_NOTNULL(regl, REGL_MAGIC);

		for (n = 0; n < regl->n; n++)
			AN(buddy_return_off_extent(dskret, &regl->arr[n]));

		AN(buddy_return_ptr_page(memret, &regl->alloc));
	}

	buddy_return(dskret);
	buddy_return(memret);
}

#ifdef BUGHUNT
/* DEBUG ONLY */
static struct regl *
regionlist_collapse(const struct regionlist *rla)
{
	struct regionlist *rl;
	const struct regl *regl;
	struct regl *r;
	size_t sz = 0;
	int8_t bits;

	VSTAILQ_FOREACH(regl, &rla->head, list)
		sz += regl->n;

	sz *= sizeof *regl->arr;
	sz += sizeof *regl;
	sz += sizeof *rl;

	bits = log2up(sz);
	r = regl_alloc_bits(rla->membuddy, bits, 0, &rl);
	AN(r);
	AN(rl);

	VSTAILQ_FOREACH(regl, &rla->head, list) {
		assert(r->space >= r->n);
		assert(r->space - r->n >= regl->n);
		memcpy(&r->arr[r->n], regl->arr, regl->n * sizeof *r->arr);
		r->n += regl->n;
	}
	return (r);
}

static inline int
buddy_off_extent_cmp(const void *xa, const void *xb)
{
	const struct buddy_off_extent *a = xa, *b = xb;

	if (a->off < b->off)
		return (-1);
	if (a->off > b->off)
		return (1);
	return (0);
}
#endif // BUGHUNT

static void
regionlist_discard(struct fellow_fd *ffd, void *ioctx,
    struct regionlist **rlp)
{
	const struct stvfe_tune *tune;
	struct regl *this, *next;
	struct regionlist *rl;

	CHECK_OBJ_NOTNULL(ffd, FELLOW_FD_MAGIC);
	tune = ffd->tune;
	CHECK_OBJ_NOTNULL(tune, STVFE_TUNE_MAGIC);

	TAKE_OBJ_NOTNULL(rl, rlp, REGIONLIST_MAGIC);

	if ((ffd->cap & FFD_CAN_ANY) == 0) {
		regionlist_free(&rl, ffd->dskbuddy);
		return;
	}

	VSTAILQ_FOREACH_SAFE(this, &rl->head, list, next) {
		CHECK_OBJ_NOTNULL(this, REGL_MAGIC);
		fellow_io_regions_discard(ffd, ioctx, this->arr, this->n,
		    tune->discard_immediate, next == NULL ? 1 : 0);
	}

	regionlist_free(&rl, ffd->dskbuddy);
}
